# BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE

import os
import ctypes
import platform

import numba
import numba.typing.ctypes_utils

if platform.system() == "Windows":
    libname = "awkward.dll"
elif platform.system() == "Darwin":
    libname = "libawkward.dylib"
else:
    libname = "libawkward.so"

libpath = os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), libname)

lib = ctypes.cdll.LoadLibrary(libpath)

# bool awkward_FillableArray_length(void* fillablearray, int64_t* result);
FillableArray_length = lib.awkward_FillableArray_length
FillableArray_length.name = "FillableArray.length"
FillableArray_length.argtypes = [ctypes.c_voidp, ctypes.POINTER(ctypes.c_int64)]
FillableArray_length.restype  = ctypes.c_uint8
FillableArray_length.numbatpe = numba.typing.ctypes_utils.make_function_type(FillableArray_length)

# bool awkward_FillableArray_clear(void* fillablearray);
FillableArray_clear = lib.awkward_FillableArray_clear
FillableArray_clear.name = "FillableArray.clear"
FillableArray_clear.argtypes = [ctypes.c_voidp]
FillableArray_clear.restype  = ctypes.c_uint8
FillableArray_clear.numbatpe = numba.typing.ctypes_utils.make_function_type(FillableArray_clear)

# bool awkward_FillableArray_null(void* fillablearray);
FillableArray_null = lib.awkward_FillableArray_null
FillableArray_null.name = "FillableArray.null"
FillableArray_null.argtypes = [ctypes.c_voidp]
FillableArray_null.restype  = ctypes.c_uint8
FillableArray_null.numbatpe = numba.typing.ctypes_utils.make_function_type(FillableArray_null)

# bool awkward_FillableArray_boolean(void* fillablearray, bool x);
FillableArray_boolean = lib.awkward_FillableArray_boolean
FillableArray_boolean.name = "FillableArray.boolean"
FillableArray_boolean.argtypes = [ctypes.c_voidp, ctypes.c_uint8]
FillableArray_boolean.restype  = ctypes.c_uint8
FillableArray_boolean.numbatpe = numba.typing.ctypes_utils.make_function_type(FillableArray_boolean)

# bool awkward_FillableArray_integer(void* fillablearray, int64_t x);
FillableArray_integer = lib.awkward_FillableArray_integer
FillableArray_integer.name = "FillableArray.integer"
FillableArray_integer.argtypes = [ctypes.c_voidp, ctypes.c_int64]
FillableArray_integer.restype  = ctypes.c_uint8
FillableArray_integer.numbatpe = numba.typing.ctypes_utils.make_function_type(FillableArray_integer)

# bool awkward_FillableArray_real(void* fillablearray, double x);
FillableArray_real = lib.awkward_FillableArray_real
FillableArray_real.name = "FillableArray.real"
FillableArray_real.argtypes = [ctypes.c_voidp, ctypes.c_double]
FillableArray_real.restype  = ctypes.c_uint8
FillableArray_real.numbatpe = numba.typing.ctypes_utils.make_function_type(FillableArray_real)

# bool awkward_FillableArray_beginlist(void* fillablearray);
FillableArray_beginlist = lib.awkward_FillableArray_beginlist
FillableArray_beginlist.name = "FillableArray.beginlist"
FillableArray_beginlist.argtypes = [ctypes.c_voidp]
FillableArray_beginlist.restype  = ctypes.c_uint8
FillableArray_beginlist.numbatpe = numba.typing.ctypes_utils.make_function_type(FillableArray_beginlist)

# bool awkward_FillableArray_endlist(void* fillablearray);
FillableArray_endlist = lib.awkward_FillableArray_endlist
FillableArray_endlist.name = "FillableArray.endlist"
FillableArray_endlist.argtypes = [ctypes.c_voidp]
FillableArray_endlist.restype  = ctypes.c_uint8
FillableArray_endlist.numbatpe = numba.typing.ctypes_utils.make_function_type(FillableArray_endlist)

# uint8_t awkward_FillableArray_begintuple(void* fillablearray, int64_t numfields);
FillableArray_begintuple = lib.awkward_FillableArray_begintuple
FillableArray_begintuple.name = "FillableArray.begintuple"
FillableArray_begintuple.argtypes = [ctypes.c_voidp, ctypes.c_int64]
FillableArray_begintuple.restype  = ctypes.c_uint8
FillableArray_begintuple.numbatpe = numba.typing.ctypes_utils.make_function_type(FillableArray_begintuple)

# uint8_t awkward_FillableArray_index(void* fillablearray, int64_t index);
FillableArray_index = lib.awkward_FillableArray_index
FillableArray_index.name = "FillableArray.index"
FillableArray_index.argtypes = [ctypes.c_voidp, ctypes.c_int64]
FillableArray_index.restype  = ctypes.c_uint8
FillableArray_index.numbatpe = numba.typing.ctypes_utils.make_function_type(FillableArray_index)

# uint8_t awkward_FillableArray_endtuple(void* fillablearray);
FillableArray_endtuple = lib.awkward_FillableArray_endtuple
FillableArray_endtuple.name = "FillableArray.endtuple"
FillableArray_endtuple.argtypes = [ctypes.c_voidp]
FillableArray_endtuple.restype  = ctypes.c_uint8
FillableArray_endtuple.numbatpe = numba.typing.ctypes_utils.make_function_type(FillableArray_endtuple)

# uint8_t awkward_FillableArray_beginrecord(void* fillablearray);
FillableArray_beginrecord = lib.awkward_FillableArray_beginrecord
FillableArray_beginrecord.name = "FillableArray.beginrecord"
FillableArray_beginrecord.argtypes = [ctypes.c_voidp]
FillableArray_beginrecord.restype  = ctypes.c_uint8
FillableArray_beginrecord.numbatpe = numba.typing.ctypes_utils.make_function_type(FillableArray_beginrecord)

# uint8_t awkward_FillableArray_beginrecord_fast(void* fillablearray, const char* name);
FillableArray_beginrecord_fast = lib.awkward_FillableArray_beginrecord_fast
FillableArray_beginrecord_fast.name = "FillableArray.beginrecord_fast"
FillableArray_beginrecord_fast.argtypes = [ctypes.c_voidp, ctypes.c_voidp]
FillableArray_beginrecord_fast.restype  = ctypes.c_uint8
FillableArray_beginrecord_fast.numbatpe = numba.typing.ctypes_utils.make_function_type(FillableArray_beginrecord_fast)

# uint8_t awkward_FillableArray_beginrecord_check(void* fillablearray, const char* name);
FillableArray_beginrecord_check = lib.awkward_FillableArray_beginrecord_check
FillableArray_beginrecord_check.name = "FillableArray.beginrecord_check"
FillableArray_beginrecord_check.argtypes = [ctypes.c_voidp, ctypes.c_voidp]
FillableArray_beginrecord_check.restype  = ctypes.c_uint8
FillableArray_beginrecord_check.numbatpe = numba.typing.ctypes_utils.make_function_type(FillableArray_beginrecord_check)

# uint8_t awkward_FillableArray_field_fast(void* fillablearray, const char* key);
FillableArray_field_fast = lib.awkward_FillableArray_field_fast
FillableArray_field_fast.name = "FillableArray.field_fast"
FillableArray_field_fast.argtypes = [ctypes.c_voidp, ctypes.c_voidp]
FillableArray_field_fast.restype  = ctypes.c_uint8
FillableArray_field_fast.numbatpe = numba.typing.ctypes_utils.make_function_type(FillableArray_field_fast)

# uint8_t awkward_FillableArray_field_check(void* fillablearray, const char* key);
FillableArray_field_check = lib.awkward_FillableArray_field_check
FillableArray_field_check.name = "FillableArray.field_check"
FillableArray_field_check.argtypes = [ctypes.c_voidp, ctypes.c_voidp]
FillableArray_field_check.restype  = ctypes.c_uint8
FillableArray_field_check.numbatpe = numba.typing.ctypes_utils.make_function_type(FillableArray_field_check)

# uint8_t awkward_FillableArray_endrecord(void* fillablearray);
FillableArray_endrecord = lib.awkward_FillableArray_endrecord
FillableArray_endrecord.name = "FillableArray.endrecord"
FillableArray_endrecord.argtypes = [ctypes.c_voidp]
FillableArray_endrecord.restype  = ctypes.c_uint8
FillableArray_endrecord.numbatpe = numba.typing.ctypes_utils.make_function_type(FillableArray_endrecord)



# BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE

import sys

import pytest
import numpy

import awkward1

def test_basic():
    content = awkward1.layout.NumpyArray(numpy.array([0.0, 1.1, 2.2, 3.3, 4.4]))

    ind = numpy.array([2, 2, 0, 3, 4], dtype=numpy.int32)
    index = awkward1.layout.Index32(ind)
    array = awkward1.layout.IndexedArray32(index, content)
    assert awkward1.tolist(array) == [2.2, 2.2, 0.0, 3.3, 4.4]
    ind[3] = 1
    assert awkward1.tolist(array) == [2.2, 2.2, 0.0, 1.1, 4.4]

    ind = numpy.array([2, 2, 0, 3, 4], dtype=numpy.uint32)
    index = awkward1.layout.IndexU32(ind)
    array = awkward1.layout.IndexedArrayU32(index, content)
    assert awkward1.tolist(array) == [2.2, 2.2, 0.0, 3.3, 4.4]
    ind[3] = 1
    assert awkward1.tolist(array) == [2.2, 2.2, 0.0, 1.1, 4.4]

    ind = numpy.array([2, 2, 0, 3, 4], dtype=numpy.int64)
    index = awkward1.layout.Index64(ind)
    array = awkward1.layout.IndexedArray64(index, content)
    assert awkward1.tolist(array) == [2.2, 2.2, 0.0, 3.3, 4.4]
    ind[3] = 1
    assert awkward1.tolist(array) == [2.2, 2.2, 0.0, 1.1, 4.4]

    ind = numpy.array([2, 2, 0, 3, 4], dtype=numpy.int32)
    index = awkward1.layout.Index32(ind)
    array = awkward1.layout.IndexedOptionArray32(index, content)
    assert awkward1.tolist(array) == [2.2, 2.2, 0.0, 3.3, 4.4]
    ind[3] = 1
    assert awkward1.tolist(array) == [2.2, 2.2, 0.0, 1.1, 4.4]

    ind = numpy.array([2, 2, 0, 3, 4], dtype=numpy.int64)
    index = awkward1.layout.Index64(ind)
    array = awkward1.layout.IndexedOptionArray64(index, content)
    assert awkward1.tolist(array) == [2.2, 2.2, 0.0, 3.3, 4.4]
    ind[3] = 1
    assert awkward1.tolist(array) == [2.2, 2.2, 0.0, 1.1, 4.4]

def test_type():
    content = awkward1.layout.NumpyArray(numpy.array([0.0, 1.1, 2.2, 3.3, 4.4]))
    index = awkward1.layout.Index32(numpy.array([2, 2, 0, 3, 4], dtype=numpy.int32))
    array = awkward1.layout.IndexedArray32(index, content)
    assert awkward1.typeof(array) == awkward1.layout.PrimitiveType("float64")
    array = awkward1.layout.IndexedOptionArray32(index, content)
    assert awkward1.typeof(array) == awkward1.layout.OptionType(awkward1.layout.PrimitiveType("float64"))

def test_null():
    content = awkward1.layout.NumpyArray(numpy.array([0.0, 1.1, 2.2, 3.3, 4.4]))
    index = awkward1.layout.Index64(numpy.array([2, 2, 0, -1, 4], dtype=numpy.int64))
    array = awkward1.layout.IndexedOptionArray64(index, content)
    assert awkward1.tolist(array) == [2.2, 2.2, 0.0, None, 4.4]

def test_carry():
    content = awkward1.layout.NumpyArray(numpy.array([0.0, 1.1, 2.2, 3.3, 4.4]))
    index = awkward1.layout.Index64(numpy.array([2, 2, 0, 3, 4], dtype=numpy.int64))
    indexedarray = awkward1.layout.IndexedArray64(index, content)
    offsets = awkward1.layout.Index64(numpy.array([0, 3, 3, 5], dtype=numpy.int64))
    listoffsetarray = awkward1.layout.ListOffsetArray64(offsets, indexedarray)
    assert awkward1.tolist(listoffsetarray) == [[2.2, 2.2, 0.0], [], [3.3, 4.4]]
    assert awkward1.tolist(listoffsetarray[::-1]) == [[3.3, 4.4], [], [2.2, 2.2, 0.0]]
    assert awkward1.tolist(listoffsetarray[[2, 0]]) == [[3.3, 4.4], [2.2, 2.2, 0.0]]
    assert awkward1.tolist(listoffsetarray[[2, 0], 1]) == [4.4, 2.2]     # invokes carry
    assert awkward1.tolist(listoffsetarray[2:, 1]) == [4.4]              # invokes carry

    index = awkward1.layout.Index64(numpy.array([2, 2, 0, 3, -1], dtype=numpy.int64))
    indexedarray = awkward1.layout.IndexedOptionArray64(index, content)
    listoffsetarray = awkward1.layout.ListOffsetArray64(offsets, indexedarray)
    assert awkward1.tolist(listoffsetarray) == [[2.2, 2.2, 0.0], [], [3.3, None]]
    assert awkward1.tolist(listoffsetarray[::-1]) == [[3.3, None], [], [2.2, 2.2, 0.0]]
    assert awkward1.tolist(listoffsetarray[[2, 0]]) == [[3.3, None], [2.2, 2.2, 0.0]]
    assert awkward1.tolist(listoffsetarray[[2, 0], 1]) == [None, 2.2]    # invokes carry
    assert awkward1.tolist(listoffsetarray[2:, 1]) == [None]             # invokes carry

def test_others():
    content = awkward1.layout.NumpyArray(numpy.array([[0.0, 0.0], [0.1, 1.0], [0.2, 2.0], [0.3, 3.0], [0.4, 4.0], [0.5, 5.0]]))
    index = awkward1.layout.Index64(numpy.array([4, 0, 3, 1, 3], dtype=numpy.int64))
    indexedarray = awkward1.layout.IndexedArray64(index, content)
    assert indexedarray[3, 0] == 0.1
    assert indexedarray[3, 1] == 1.0
    assert awkward1.tolist(indexedarray[3, ::-1]) == [1.0, 0.1]
    assert awkward1.tolist(indexedarray[3, [1, 1, 0]]) == [1.0, 1.0, 0.1]
    assert awkward1.tolist(indexedarray[3:, 0]) == [0.1, 0.3]
    assert awkward1.tolist(indexedarray[3:, 1]) == [1.0, 3.0]
    assert awkward1.tolist(indexedarray[3:, ::-1]) == [[1.0, 0.1], [3.0, 0.3]]
    assert awkward1.tolist(indexedarray[3:, [1, 1, 0]]) == [[1.0, 1.0, 0.1], [3.0, 3.0, 0.3]]

def test_missing():
    content = awkward1.layout.NumpyArray(numpy.array([[0.0, 0.0], [0.1, 1.0], [0.2, 2.0], [0.3, 3.0], [0.4, 4.0], [0.5, 5.0]]))
    index = awkward1.layout.Index64(numpy.array([4, 0, 3, -1, 3], dtype=numpy.int64))
    indexedarray = awkward1.layout.IndexedOptionArray64(index, content)
    assert awkward1.tolist(indexedarray[3:, 0]) == [None, 0.3]
    assert awkward1.tolist(indexedarray[3:, 1]) == [None, 3.0]
    assert awkward1.tolist(indexedarray[3:, ::-1]) == [None, [3.0, 0.3]]
    assert awkward1.tolist(indexedarray[3:, [1, 1, 0]]) == [None, [3.0, 3.0, 0.3]]

def test_highlevel():
    content = awkward1.layout.NumpyArray(numpy.array([0.0, 1.1, 2.2, 3.3, 4.4]))
    index = awkward1.layout.Index64(numpy.array([2, 2, 0, -1, 4], dtype=numpy.int64))
    array = awkward1.Array(awkward1.layout.IndexedOptionArray64(index, content))
    assert awkward1.tolist(array) == [2.2, 2.2, 0.0, None, 4.4]
    assert str(array) == "[2.2, 2.2, 0, None, 4.4]"
    assert repr(array) == "<Array [2.2, 2.2, 0, None, 4.4] type='5 * ?float64'>"

def test_fillable():
    assert awkward1.tolist(awkward1.Array([1.1, 2.2, 3.3, None, 4.4])) == [1.1, 2.2, 3.3, None, 4.4]
    assert awkward1.tolist(awkward1.Array([None, 2.2, 3.3, None, 4.4])) == [None, 2.2, 3.3, None, 4.4]

    assert awkward1.tolist(awkward1.Array([[1.1, 2.2, 3.3], [], [None, 4.4]])) == [[1.1, 2.2, 3.3], [], [None, 4.4]]
    assert awkward1.tolist(awkward1.Array([[1.1, 2.2, 3.3], [], None, [None, 4.4]])) == [[1.1, 2.2, 3.3], [], None, [None, 4.4]]
    assert awkward1.tolist(awkward1.Array([[1.1, 2.2, 3.3], None, [], [None, 4.4]])) == [[1.1, 2.2, 3.3], None, [], [None, 4.4]]
    assert awkward1.tolist(awkward1.Array([[1.1, 2.2, 3.3], None, [], [None, 4.4]])) != [[1.1, 2.2, 3.3], [], None, [None, 4.4]]

    assert awkward1.tolist(awkward1.Array([[None, 1.1, 2.2, 3.3], [], [None, 4.4]])) == [[None, 1.1, 2.2, 3.3], [], [None, 4.4]]
    assert awkward1.tolist(awkward1.Array([[None, 1.1, 2.2, 3.3], [], None, [None, 4.4]])) == [[None, 1.1, 2.2, 3.3], [], None, [None, 4.4]]
    assert awkward1.tolist(awkward1.Array([[None, 1.1, 2.2, 3.3], None, [], [None, 4.4]])) == [[None, 1.1, 2.2, 3.3], None, [], [None, 4.4]]
    assert awkward1.tolist(awkward1.Array([[None, 1.1, 2.2, 3.3], None, [], [None, 4.4]])) != [[None, 1.1, 2.2, 3.3], [], None, [None, 4.4]]

    assert awkward1.tolist(awkward1.Array([None, [1.1, 2.2, 3.3], [], [None, 4.4]])) == [None, [1.1, 2.2, 3.3], [], [None, 4.4]]
    assert awkward1.tolist(awkward1.Array([None, [1.1, 2.2, 3.3], [], None, [None, 4.4]])) == [None, [1.1, 2.2, 3.3], [], None, [None, 4.4]]
    assert awkward1.tolist(awkward1.Array([None, [1.1, 2.2, 3.3], None, [], [None, 4.4]])) == [None, [1.1, 2.2, 3.3], None, [], [None, 4.4]]
    assert awkward1.tolist(awkward1.Array([None, [1.1, 2.2, 3.3], None, [], [None, 4.4]])) != [None, [1.1, 2.2, 3.3], [], None, [None, 4.4]]

    assert awkward1.tolist(awkward1.Array([None, None, None, None, None])) == [None, None, None, None, None]
    assert awkward1.tolist(awkward1.Array([[None, None, None], [], [None, None]])) == [[None, None, None], [], [None, None]]

def test_json():
    assert awkward1.tolist(awkward1.Array("[1.1, 2.2, 3.3, null, 4.4]")) == [1.1, 2.2, 3.3, None, 4.4]
    assert awkward1.tolist(awkward1.Array("[null, 2.2, 3.3, null, 4.4]")) == [None, 2.2, 3.3, None, 4.4]

    assert awkward1.tolist(awkward1.Array("[[1.1, 2.2, 3.3], [], [null, 4.4]]")) == [[1.1, 2.2, 3.3], [], [None, 4.4]]
    assert awkward1.tolist(awkward1.Array("[[1.1, 2.2, 3.3], [], null, [null, 4.4]]")) == [[1.1, 2.2, 3.3], [], None, [None, 4.4]]
    assert awkward1.tolist(awkward1.Array("[[1.1, 2.2, 3.3], null, [], [null, 4.4]]")) == [[1.1, 2.2, 3.3], None, [], [None, 4.4]]
    assert awkward1.tolist(awkward1.Array("[[1.1, 2.2, 3.3], null, [], [null, 4.4]]")) != [[1.1, 2.2, 3.3], [], None, [None, 4.4]]

    assert awkward1.tolist(awkward1.Array("[[null, 1.1, 2.2, 3.3], [], [null, 4.4]]")) == [[None, 1.1, 2.2, 3.3], [], [None, 4.4]]
    assert awkward1.tolist(awkward1.Array("[[null, 1.1, 2.2, 3.3], [], null, [null, 4.4]]")) == [[None, 1.1, 2.2, 3.3], [], None, [None, 4.4]]
    assert awkward1.tolist(awkward1.Array("[[null, 1.1, 2.2, 3.3], null, [], [null, 4.4]]")) == [[None, 1.1, 2.2, 3.3], None, [], [None, 4.4]]
    assert awkward1.tolist(awkward1.Array("[[null, 1.1, 2.2, 3.3], null, [], [null, 4.4]]")) != [[None, 1.1, 2.2, 3.3], [], None, [None, 4.4]]

    assert awkward1.tolist(awkward1.Array("[null, [1.1, 2.2, 3.3], [], [null, 4.4]]")) == [None, [1.1, 2.2, 3.3], [], [None, 4.4]]
    assert awkward1.tolist(awkward1.Array("[null, [1.1, 2.2, 3.3], [], null, [null, 4.4]]")) == [None, [1.1, 2.2, 3.3], [], None, [None, 4.4]]
    assert awkward1.tolist(awkward1.Array("[null, [1.1, 2.2, 3.3], null, [], [null, 4.4]]")) == [None, [1.1, 2.2, 3.3], None, [], [None, 4.4]]
    assert awkward1.tolist(awkward1.Array("[null, [1.1, 2.2, 3.3], null, [], [null, 4.4]]")) != [None, [1.1, 2.2, 3.3], [], None, [None, 4.4]]

    assert awkward1.tolist(awkward1.Array("[null, null, null, null, null]")) == [None, None, None, None, None]
    assert awkward1.tolist(awkward1.Array("[[null, null, null], [], [null, null]]")) == [[None, None, None], [], [None, None]]

def test_emptyarray():
    array = awkward1.layout.EmptyArray()
    assert isinstance(array.astype(awkward1.layout.OptionType(awkward1.layout.PrimitiveType("float32"))), awkward1.layout.IndexedOptionArray64)

    offsets = awkward1.layout.Index64(numpy.array([0], dtype=numpy.int64))
    array = awkward1.layout.ListOffsetArray64(offsets, awkward1.layout.EmptyArray())
    assert isinstance(array.astype(awkward1.layout.ListType(awkward1.layout.OptionType(awkward1.layout.PrimitiveType("float32")))).content, awkward1.layout.IndexedOptionArray64)



# BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE

import sys
import itertools

import pytest
import numpy

import awkward1

py27 = (sys.version_info[0] < 3)

content = awkward1.layout.NumpyArray(numpy.arange(2*3*5*7).reshape(-1, 7))
offsetsA = numpy.arange(0, 2*3*5 + 5, 5)
offsetsB = numpy.arange(0, 2*3 + 3, 3)
startsA, stopsA = offsetsA[:-1], offsetsA[1:]
startsB, stopsB = offsetsB[:-1], offsetsB[1:]

listoffsetarrayA64 = awkward1.layout.ListOffsetArray64(awkward1.layout.Index64(offsetsA), content)
listoffsetarrayA32 = awkward1.layout.ListOffsetArray32(awkward1.layout.Index32(offsetsA), content)
listarrayA64 = awkward1.layout.ListArray64(awkward1.layout.Index64(startsA), awkward1.layout.Index64(stopsA), content)
listarrayA32 = awkward1.layout.ListArray32(awkward1.layout.Index32(startsA), awkward1.layout.Index32(stopsA), content)
modelA = numpy.arange(2*3*5*7).reshape(2*3, 5, 7)

listoffsetarrayB64 = awkward1.layout.ListOffsetArray64(awkward1.layout.Index64(offsetsB), listoffsetarrayA64)
listoffsetarrayB32 = awkward1.layout.ListOffsetArray32(awkward1.layout.Index32(offsetsB), listoffsetarrayA64)
listarrayB64 = awkward1.layout.ListArray64(awkward1.layout.Index64(startsB), awkward1.layout.Index64(stopsB), listarrayA64)
listarrayB32 = awkward1.layout.ListArray32(awkward1.layout.Index32(startsB), awkward1.layout.Index32(stopsB), listarrayA64)
modelB = numpy.arange(2*3*5*7).reshape(2, 3, 5, 7)

listoffsetarrayB64.setidentities()
listoffsetarrayB32.setidentities()
listarrayB64.setidentities()
listarrayB32.setidentities()

def test_basic():
    assert awkward1.tolist(modelA) == awkward1.tolist(listoffsetarrayA64)
    assert awkward1.tolist(modelA) == awkward1.tolist(listoffsetarrayA32)
    assert awkward1.tolist(modelA) == awkward1.tolist(listarrayA64)
    assert awkward1.tolist(modelA) == awkward1.tolist(listarrayA32)
    assert awkward1.tolist(modelB) == awkward1.tolist(listoffsetarrayB64)
    assert awkward1.tolist(modelB) == awkward1.tolist(listoffsetarrayB32)
    assert awkward1.tolist(modelB) == awkward1.tolist(listarrayB64)
    assert awkward1.tolist(modelB) == awkward1.tolist(listarrayB32)

def test_listoffsetarrayA64():
    for depth in 0, 1, 2, 3:
        for cuts in itertools.permutations((0, 1, 4, -5), depth):
            assert awkward1.tolist(modelA[cuts]) == awkward1.tolist(listoffsetarrayA64[cuts])

    for depth in 0, 1, 2, 3:
        for cuts in itertools.permutations((slice(None), slice(1, None), slice(None, -1), slice(None, None, 2)), depth):
            assert awkward1.tolist(modelA[cuts]) == awkward1.tolist(listoffsetarrayA64[cuts])

    for depth in 0, 1, 2, 3:
        for cuts in itertools.permutations((slice(1, None), slice(None, -1), 2, -2), depth):
            assert awkward1.tolist(modelA[cuts]) == awkward1.tolist(listoffsetarrayA64[cuts])

    for depth in 0, 1, 2, 3:
        for cuts in itertools.permutations(([2, 0, 0, 1], [1, -2, 0, -1], 2, -2), depth):
            assert awkward1.tolist(modelA[cuts]) == awkward1.tolist(listoffsetarrayA64[cuts])

    for depth in 0, 1, 2, 3:
        for cuts in itertools.permutations(([2, 0, 0, 1], [1, -2, 0, -1], slice(1, None), slice(None, -1)), depth):
            cuts = cuts
            while len(cuts) > 0 and isinstance(cuts[0], slice):
                cuts = cuts[1:]
            while len(cuts) > 0 and isinstance(cuts[-1], slice):
                cuts = cuts[:-1]
            if any(isinstance(x, slice) for x in cuts):
                continue
            assert awkward1.tolist(modelA[cuts]) == awkward1.tolist(listoffsetarrayA64[cuts])

def test_listoffsetarrayA32():
    for depth in 0, 1, 2, 3:
        for cuts in itertools.permutations((0, 1, 4, -5), depth):
            assert awkward1.tolist(modelA[cuts]) == awkward1.tolist(listoffsetarrayA32[cuts])

    for depth in 0, 1, 2, 3:
        for cuts in itertools.permutations((slice(None), slice(1, None), slice(None, -1), slice(None, None, 2)), depth):
            assert awkward1.tolist(modelA[cuts]) == awkward1.tolist(listoffsetarrayA32[cuts])

    for depth in 0, 1, 2, 3:
        for cuts in itertools.permutations((slice(1, None), slice(None, -1), 2, -2), depth):
            assert awkward1.tolist(modelA[cuts]) == awkward1.tolist(listoffsetarrayA32[cuts])

    for depth in 0, 1, 2, 3:
        for cuts in itertools.permutations(([2, 0, 0, 1], [1, -2, 0, -1], 2, -2), depth):
            assert awkward1.tolist(modelA[cuts]) == awkward1.tolist(listoffsetarrayA32[cuts])

    for depth in 0, 1, 2, 3:
        for cuts in itertools.permutations(([2, 0, 0, 1], [1, -2, 0, -1], slice(1, None), slice(None, -1)), depth):
            cuts = cuts
            while len(cuts) > 0 and isinstance(cuts[0], slice):
                cuts = cuts[1:]
            while len(cuts) > 0 and isinstance(cuts[-1], slice):
                cuts = cuts[:-1]
            if any(isinstance(x, slice) for x in cuts):
                continue
            assert awkward1.tolist(modelA[cuts]) == awkward1.tolist(listoffsetarrayA32[cuts])

def test_listarrayA64():
    for depth in 0, 1, 2, 3:
        for cuts in itertools.permutations((0, 1, 4, -5), depth):
            assert awkward1.tolist(modelA[cuts]) == awkward1.tolist(listarrayA64[cuts])

    for depth in 0, 1, 2, 3:
        for cuts in itertools.permutations((slice(None), slice(1, None), slice(None, -1), slice(None, None, 2)), depth):
            assert awkward1.tolist(modelA[cuts]) == awkward1.tolist(listarrayA64[cuts])

    for depth in 0, 1, 2, 3:
        for cuts in itertools.permutations((slice(1, None), slice(None, -1), 2, -2), depth):
            assert awkward1.tolist(modelA[cuts]) == awkward1.tolist(listarrayA64[cuts])

    for depth in 0, 1, 2, 3:
        for cuts in itertools.permutations(([2, 0, 0, 1], [1, -2, 0, -1], 2, -2), depth):
            assert awkward1.tolist(modelA[cuts]) == awkward1.tolist(listarrayA64[cuts])

    for depth in 0, 1, 2, 3:
        for cuts in itertools.permutations(([2, 0, 0, 1], [1, -2, 0, -1], slice(1, None), slice(None, -1)), depth):
            cuts = cuts
            while len(cuts) > 0 and isinstance(cuts[0], slice):
                cuts = cuts[1:]
            while len(cuts) > 0 and isinstance(cuts[-1], slice):
                cuts = cuts[:-1]
            if any(isinstance(x, slice) for x in cuts):
                continue
            assert awkward1.tolist(modelA[cuts]) == awkward1.tolist(listarrayA64[cuts])

def test_listarrayA32():
    for depth in 0, 1, 2, 3:
        for cuts in itertools.permutations((0, 1, 4, -5), depth):
            assert awkward1.tolist(modelA[cuts]) == awkward1.tolist(listarrayA32[cuts])

    for depth in 0, 1, 2, 3:
        for cuts in itertools.permutations((slice(None), slice(1, None), slice(None, -1), slice(None, None, 2)), depth):
            assert awkward1.tolist(modelA[cuts]) == awkward1.tolist(listarrayA32[cuts])

    for depth in 0, 1, 2, 3:
        for cuts in itertools.permutations((slice(1, None), slice(None, -1), 2, -2), depth):
            assert awkward1.tolist(modelA[cuts]) == awkward1.tolist(listarrayA32[cuts])

    for depth in 0, 1, 2, 3:
        for cuts in itertools.permutations(([2, 0, 0, 1], [1, -2, 0, -1], 2, -2), depth):
            assert awkward1.tolist(modelA[cuts]) == awkward1.tolist(listarrayA32[cuts])

    for depth in 0, 1, 2, 3:
        for cuts in itertools.permutations(([2, 0, 0, 1], [1, -2, 0, -1], slice(1, None), slice(None, -1)), depth):
            cuts = cuts
            while len(cuts) > 0 and isinstance(cuts[0], slice):
                cuts = cuts[1:]
            while len(cuts) > 0 and isinstance(cuts[-1], slice):
                cuts = cuts[:-1]
            if any(isinstance(x, slice) for x in cuts):
                continue
            assert awkward1.tolist(modelA[cuts]) == awkward1.tolist(listarrayA32[cuts])

def test_listoffsetarrayB64():
    for depth in 0, 1, 2, 3, 4:
        for cuts in itertools.permutations((-2, -1, 0, 1, 1), depth):
            assert awkward1.tolist(modelB[cuts]) == awkward1.tolist(listoffsetarrayB64[cuts])

    for depth in 0, 1, 2, 3, 4:
        for cuts in itertools.permutations((-1, 0, 1, slice(1, None), slice(None, -1)), depth):
            assert awkward1.tolist(modelB[cuts]) == awkward1.tolist(listoffsetarrayB64[cuts])

    for depth in 0, 1, 2, 3, 4:
        for cuts in itertools.permutations((-1, 0, [1, 0, 0, 1], [0, 1, -1, 1], slice(None, -1)), depth):
            cuts = cuts
            while len(cuts) > 0 and isinstance(cuts[0], slice):
                cuts = cuts[1:]
            while len(cuts) > 0 and isinstance(cuts[-1], slice):
                cuts = cuts[:-1]
            if any(isinstance(x, slice) for x in cuts):
                continue
            assert awkward1.tolist(modelB[cuts]) == awkward1.tolist(listoffsetarrayB64[cuts])

def test_listoffsetarrayB32():
    for depth in 0, 1, 2, 3, 4:
        for cuts in itertools.permutations((-2, -1, 0, 1, 1), depth):
            assert awkward1.tolist(modelB[cuts]) == awkward1.tolist(listoffsetarrayB64[cuts])

    for depth in 0, 1, 2, 3, 4:
        for cuts in itertools.permutations((-1, 0, 1, slice(1, None), slice(None, -1)), depth):
            assert awkward1.tolist(modelB[cuts]) == awkward1.tolist(listoffsetarrayB64[cuts])

    for depth in 0, 1, 2, 3, 4:
        for cuts in itertools.permutations((-1, 0, [1, 0, 0, 1], [0, 1, -1, 1], slice(None, -1)), depth):
            cuts = cuts
            while len(cuts) > 0 and isinstance(cuts[0], slice):
                cuts = cuts[1:]
            while len(cuts) > 0 and isinstance(cuts[-1], slice):
                cuts = cuts[:-1]
            if any(isinstance(x, slice) for x in cuts):
                continue
            assert awkward1.tolist(modelB[cuts]) == awkward1.tolist(listoffsetarrayB64[cuts])

def test_listarrayB64():
    for depth in 0, 1, 2, 3, 4:
        for cuts in itertools.permutations((-2, -1, 0, 1, 1), depth):
            assert awkward1.tolist(modelB[cuts]) == awkward1.tolist(listarrayB64[cuts])

    for depth in 0, 1, 2, 3, 4:
        for cuts in itertools.permutations((-1, 0, 1, slice(1, None), slice(None, -1)), depth):
            assert awkward1.tolist(modelB[cuts]) == awkward1.tolist(listarrayB64[cuts])

    for depth in 0, 1, 2, 3, 4:
        for cuts in itertools.permutations((-1, 0, [1, 0, 0, 1], [0, 1, -1, 1], slice(None, -1)), depth):
            cuts = cuts
            while len(cuts) > 0 and isinstance(cuts[0], slice):
                cuts = cuts[1:]
            while len(cuts) > 0 and isinstance(cuts[-1], slice):
                cuts = cuts[:-1]
            if any(isinstance(x, slice) for x in cuts):
                continue
            assert awkward1.tolist(modelB[cuts]) == awkward1.tolist(listarrayB64[cuts])

def test_listarrayB32():
    for depth in 0, 1, 2, 3, 4:
        for cuts in itertools.permutations((-2, -1, 0, 1, 1), depth):
            assert awkward1.tolist(modelB[cuts]) == awkward1.tolist(listarrayB64[cuts])

    for depth in 0, 1, 2, 3, 4:
        for cuts in itertools.permutations((-1, 0, 1, slice(1, None), slice(None, -1)), depth):
            assert awkward1.tolist(modelB[cuts]) == awkward1.tolist(listarrayB64[cuts])

    for depth in 0, 1, 2, 3, 4:
        for cuts in itertools.permutations((-1, 0, [1, 0, 0, 1], [0, 1, -1, 1], slice(None, -1)), depth):
            cuts = cuts
            while len(cuts) > 0 and isinstance(cuts[0], slice):
                cuts = cuts[1:]
            while len(cuts) > 0 and isinstance(cuts[-1], slice):
                cuts = cuts[:-1]
            if any(isinstance(x, slice) for x in cuts):
                continue
            assert awkward1.tolist(modelB[cuts]) == awkward1.tolist(listarrayB64[cuts])



# BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE

import sys
import os
import json

import pytest
import numpy

import awkward1

numba = pytest.importorskip("numba")

def test_boxing():
    @numba.njit
    def f1(q):
        z = q
        return 3.14

    a = awkward1.layout.FillableArray()
    assert sys.getrefcount(a) == 2
    f1(a)
    assert sys.getrefcount(a) == 2

    @numba.njit
    def f2(q):
        z = q
        return q

    a = awkward1.layout.FillableArray()
    assert sys.getrefcount(a) == 2
    f2(a)
    assert sys.getrefcount(a) == 2
    b = f2(a)
    assert sys.getrefcount(a) == 3

    assert str(b.snapshot()) == """<EmptyArray/>"""

def test_simple():
    @numba.njit
    def f1(q):
        q.clear()
        return 3.14

    a = awkward1.layout.FillableArray()
    f1(a)

def test_boolean():
    @numba.njit
    def f1(q):
        q.boolean(True)
        q.boolean(False)
        q.boolean(False)
        return q

    a = awkward1.layout.FillableArray()
    b = f1(a)
    assert awkward1.tolist(a.snapshot()) == [True, False, False]
    assert awkward1.tolist(b.snapshot()) == [True, False, False]

def test_integer():
    @numba.njit
    def f1(q):
        q.integer(1)
        q.integer(2)
        q.integer(3)
        return q

    a = awkward1.layout.FillableArray()
    b = f1(a)
    assert awkward1.tolist(a.snapshot()) == [1, 2, 3]
    assert awkward1.tolist(b.snapshot()) == [1, 2, 3]

def test_real():
    @numba.njit
    def f1(q, z):
        q.real(1)
        q.real(2.2)
        q.real(z)
        return q

    a = awkward1.layout.FillableArray()
    b = f1(a, numpy.array([3.5], dtype=numpy.float32)[0])
    assert awkward1.tolist(a.snapshot()) == [1, 2.2, 3.5]
    assert awkward1.tolist(b.snapshot()) == [1, 2.2, 3.5]

def test_list():
    @numba.njit
    def f1(q):
        q.beginlist()
        q.real(1.1)
        q.real(2.2)
        q.real(3.3)
        q.endlist()
        q.beginlist()
        q.endlist()
        q.beginlist()
        q.real(4.4)
        q.real(5.5)
        q.endlist()
        return q

    a = awkward1.layout.FillableArray()
    b = f1(a)
    assert awkward1.tolist(a.snapshot()) == [[1.1, 2.2, 3.3], [], [4.4, 5.5]]
    assert awkward1.tolist(b.snapshot()) == [[1.1, 2.2, 3.3], [], [4.4, 5.5]]

    @numba.njit
    def f2(q):
        return len(q)

    assert f2(a) == 3
    assert f2(b) == 3

    @numba.njit
    def f3(q):
        q.clear()
        return q

    c = f3(b)
    assert awkward1.tolist(a.snapshot()) == []
    assert awkward1.tolist(b.snapshot()) == []
    assert awkward1.tolist(c.snapshot()) == []



# BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE

import sys
import gc

import pytest
import numpy
numba = pytest.importorskip("numba")

import awkward1

py27 = 2 if sys.version_info[0] < 3 else 1

def test_reminder():
    x1 = numpy.arange(10, dtype="i4")
    x2 = awkward1.layout.NumpyArray(x1)
    assert (sys.getrefcount(x1), sys.getrefcount(x2)) == (3, 2)

    @numba.njit
    def f1(q):
        pass
    f1(x2)
    assert (sys.getrefcount(x1), sys.getrefcount(x2)) == (3, 2)

    @numba.njit
    def f2(q):
        return q
    tmp = f2(x2)
    assert (sys.getrefcount(x1), sys.getrefcount(x2)) == (3, 2 + 1*py27)

    del tmp
    assert (sys.getrefcount(x1), sys.getrefcount(x2)) == (3, 2)

    del x2
    assert sys.getrefcount(x1) == 2

def test_refcount():
    array = numpy.arange(40, dtype="i4").reshape(-1, 4)
    identity = awkward1.layout.Identities32(0, [], array)
    assert (sys.getrefcount(array), sys.getrefcount(identity)) == (3, 2)

    @numba.njit
    def f1(q):
        pass
    f1(identity)
    assert (sys.getrefcount(array), sys.getrefcount(identity)) == (3, 2)

    @numba.njit
    def f2(q):
        return q
    tmp = f2(identity)
    assert (sys.getrefcount(array), sys.getrefcount(identity)) == (3, 2 + 1*py27)

    del tmp
    assert (sys.getrefcount(array), sys.getrefcount(identity)) == (3, 2)

    tmp = f2(identity)
    assert (sys.getrefcount(array), sys.getrefcount(identity)) == (3, 2 + 1*py27)

    del array
    del identity
    del tmp

def test_width():
    identity = awkward1.layout.Identities32(0, [], numpy.arange(40, dtype="i4").reshape(-1, 4))
    @numba.njit
    def f1(q):
        return q.width
    assert f1(identity) == 4

def test_numpyarray():
    i1 = numpy.arange(3, dtype="i4").reshape(-1, 1)
    i2 = awkward1.layout.Identities32(0, [], i1)
    content = awkward1.layout.NumpyArray(numpy.array([1.1, 2.2, 3.3]), identities=i2)

    assert (sys.getrefcount(i1), sys.getrefcount(i2), sys.getrefcount(content)) == (3, 2, 2)

    @numba.njit
    def f1(q):
        return q

    tmp = f1(content)
    assert (sys.getrefcount(i1), sys.getrefcount(i2), sys.getrefcount(content)) == (3, 2, 2 + 1*py27)

    tmp2 = tmp.identities
    assert (sys.getrefcount(i1), sys.getrefcount(i2), sys.getrefcount(content)) == (3, 2, 2 + 1*py27)

    del tmp
    assert (sys.getrefcount(i1), sys.getrefcount(i2), sys.getrefcount(content)) == (3, 2, 2)

def test_listoffsetarray():
    i1 = numpy.arange(3, dtype="i4").reshape(-1, 1)
    i2 = awkward1.layout.Identities32(0, [], i1)
    offsets = awkward1.layout.Index32(numpy.array([0, 2, 2, 3], "i4"))
    content = awkward1.layout.NumpyArray(numpy.array([1.1, 2.2, 3.3]))
    array = awkward1.layout.ListOffsetArray32(offsets, content, identities=i2)

    assert (sys.getrefcount(i1), sys.getrefcount(i2), sys.getrefcount(array)) == (3, 2, 2)

    @numba.njit
    def f1(q):
        return q

    tmp = f1(array)
    assert (sys.getrefcount(i1), sys.getrefcount(i2), sys.getrefcount(array)) == (3, 2, 2)

    tmp2 = tmp.identities
    assert (sys.getrefcount(i1), sys.getrefcount(i2), sys.getrefcount(array)) == (3, 2, 2)

    del tmp
    assert (sys.getrefcount(i1), sys.getrefcount(i2), sys.getrefcount(array)) == (3, 2, 2)

def test_id_attribute():
    content = awkward1.layout.NumpyArray(numpy.array([1.1, 2.2, 3.3]))
    content.setidentities()
    assert numpy.asarray(content.identities).tolist() == [[0], [1], [2]]

    @numba.njit
    def f1(q):
        return q.identities

    assert numpy.asarray(f1(content)).tolist() == [[0], [1], [2]]

    offsets = awkward1.layout.Index32(numpy.array([0, 2, 2, 3], "i4"))
    array = awkward1.layout.ListOffsetArray32(offsets, content)
    array.setidentities()
    assert numpy.asarray(array.identities).tolist() == [[0], [1], [2]]
    assert numpy.asarray(array.content.identities).tolist() == [[0, 0], [0, 1], [2, 0]]

    @numba.njit
    def f2(q):
        return q.identities

    assert numpy.asarray(f2(array)).tolist() == [[0], [1], [2]]

def test_other_attributes():
    content = awkward1.layout.NumpyArray(numpy.array([1.1, 2.2, 3.3]))
    offsets = awkward1.layout.Index32(numpy.array([0, 2, 2, 3], "i4"))
    array = awkward1.layout.ListOffsetArray32(offsets, content)

    @numba.njit
    def f1(q):
        return q.content

    assert numpy.asarray(f1(array)).tolist() == [1.1, 2.2, 3.3]

    @numba.njit
    def f2(q):
        return q.offsets

    assert numpy.asarray(f2(array)).tolist() == [0, 2, 2, 3]



# BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE

import numpy

import awkward1._numpy
import awkward1._pandas
import awkward1.layout
import awkward1.operations.convert

class Array(awkward1._numpy.NDArrayOperatorsMixin, awkward1._pandas.PandasMixin):
    def __init__(self, data, type=None, classes=None, functions=None):
        if isinstance(data, awkward1.layout.Content):
            layout = data
        elif isinstance(data, Array):
            layout = data.layout
        elif isinstance(data, numpy.ndarray):
            layout = awkward1.operations.convert.fromnumpy(data).layout
        elif isinstance(data, str):
            layout = awkward1.operations.convert.fromjson(data).layout
        else:
            layout = awkward1.operations.convert.fromiter(data).layout
        if not isinstance(layout, awkward1.layout.Content):
            raise TypeError("could not convert data into an awkward1.Array")

        self._classes = classes
        self._functions = functions

        if type is not None:
            if not isinstance(type, awkward1.layout.ArrayType):
                raise TypeError("type must be an awkward1.layout.ArrayType")
            if type.length > len(layout):
                raise TypeError("ArrayType length ({0}) is greater than layout length {1}".format(type.length, len(layout)))
            if type.length < len(layout):
                layout = layout[:type.length]
            cls = awkward1._util.regular_classes(self._classes).get(type.parameters.get("__class__"))
            if cls is not None:
                if not isinstance(cls, __builtins__["type"]) or not issubclass(cls, Array):
                    raise TypeError("type.parameters['__class__'] = {0} must be a subclass of awkward1.Array".format(repr(type.parameters["__class__"])))
                self.__class__ = cls
            layout = layout.astype(type.type)

        self.layout = layout

    @property
    def layout(self):
        return self._layout

    @layout.setter
    def layout(self, layout):
        if not isinstance(layout, awkward1.layout.Content):
            raise TypeError("layout must be a subclass of awkward1.layout.Content")
        self._layout = layout

    @property
    def type(self):
        return awkward1.layout.ArrayType(self._layout.type, len(self._layout))

    def __len__(self):
        return len(self._layout)

    def __iter__(self):
        for x in self._layout:
            yield awkward1._util.wrap(x, self._classes, self._functions)

    def __getitem__(self, where):
        return awkward1._util.wrap(self._layout[where], self._classes, self._functions)

    def __str__(self, limit_value=85):
        return awkward1._util.minimally_touching_string(limit_value, self._layout, self._classes, self._functions)

    def __repr__(self, limit_value=40, limit_total=85):
        value = awkward1._util.minimally_touching_string(limit_value, self._layout, self._classes, self._functions)

        limit_type = limit_total - len(value) - len("<Array  type=>")
        type = repr(str(self.type))
        if len(type) > limit_type:
            type = type[:(limit_type - 4)] + "..." + type[-1]

        return "<Array {0} type={1}>".format(value, type)

    def __array__(self, *args, **kwargs):
        if awkward1._util.called_by_module("pandas"):
            try:
                return awkward1._numpy.convert_to_array(self._layout, args, kwargs)
            except:
                out = numpy.empty(len(self._layout), dtype="O")
                for i, x in enumerate(self._layout):
                    out[i] = awkward1._util.wrap(x, self._classes, self._functions)
                return out
        else:
            return awkward1._numpy.convert_to_array(self._layout, args, kwargs)

    def __array_function__(self, func, types, args, kwargs):
        return awkward1._numpy.array_function(func, types, args, kwargs)

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        return awkward1._numpy.array_ufunc(ufunc, method, inputs, kwargs, self._classes, self._functions)

class Record(awkward1._numpy.NDArrayOperatorsMixin):
    def __init__(self, data, type=None, classes=None, functions=None):
        # FIXME: more checks here
        layout = data
        if not isinstance(layout, awkward1.layout.Record):
            raise TypeError("could not convert data into an awkward1.Record")

        self._classes = classes
        self._functions = functions

        if type is not None:
            if not isinstance(type, awkward1.layout.RecordType):
                raise TypeError("type must be an awkward1.layout.RecordType")
            cls = awkward1._util.regular_classes(self._classes).get(type.parameters.get("__class__"))
            if cls is not None:
                if not isinstance(cls, __builtins__["type"]) or not issubclass(cls, Record):
                    raise TypeError("type.parameters['__class__'] = {0} must be a subclass of awkward1.Record".format(repr(type.parameters["__class__"])))
                self.__class__ = cls
            layout = layout.astype(type)

        self.layout = layout

    @property
    def layout(self):
        return self._layout

    @layout.setter
    def layout(self, layout):
        if not isinstance(layout, awkward1.layout.Record):
            raise TypeError("layout must be a subclass of awkward1.layout.Record")
        self._layout = layout

    @property
    def type(self):
        return self._layout.type

    def __getitem__(self, where):
        return awkward1._util.wrap(self._layout[where], self._classes, self._functions)

    def __str__(self, limit_value=85):
        return awkward1._util.minimally_touching_string(limit_value + 2, self._layout, self._classes, self._functions)[1:-1]

    def __repr__(self, limit_value=40, limit_total=85):
        value = awkward1._util.minimally_touching_string(limit_value + 2, self._layout, self._classes, self._functions)[1:-1]

        limit_type = limit_total - len(value) - len("<Record  type=>")
        type = repr(str(self._layout.type))
        if len(type) > limit_type:
            type = type[:(limit_type - 4)] + "..." + type[-1]

        return "<Record {0} type={1}>".format(value, type)

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        return awkward1._numpy.array_ufunc(ufunc, method, inputs, kwargs, self._classes, self._functions)

class FillableArray(object):
    def __init__(self, classes=None, functions=None):
        self._fillablearray = awkward1.layout.FillableArray()
        self._classes = classes
        self._functions = functions

    @property
    def type(self):
        tmp = self._fillablearray.snapshot()
        return awkward1.layout.ArrayType(tmp.type, len(tmp))

    def __len__(self):
        return len(self._fillablearray)

    def __getitem__(self, where):
        return awkward1._util.wrap(self._fillablearray[where], self._classes, self._functions)

    def __iter__(self):
        for x in self._fillablearray.snapshot():
            yield awkward1._util.wrap(x, self._classes, self._functions)

    def __str__(self, limit_value=85, snapshot=None):
        if snapshot is None:
            snapshot = self.snapshot()
        return snapshot.__str__(limit_value=limit_value)

    def __repr__(self, limit_value=40, limit_total=85):
        snapshot = self.snapshot()
        value = self.__str__(limit_value=limit_value, snapshot=snapshot)

        limit_type = limit_total - len(value) - len("<FillableArray  type=>")
        type = repr(str(snapshot.type))
        if len(type) > limit_type:
            type = type[:(limit_type - 4)] + "..." + type[-1]

        return "<FillableArray {0} type={1}>".format(value, type)

    def __array__(self, *args, **kwargs):
        return awkward1._numpy.convert_to_array(self._fillablearray.snapshot(), args, kwargs)

    def __array_function__(self, func, types, args, kwargs):
        return awkward1._numpy.array_function(func, types, args, kwargs)

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        return awkward1._numpy.array_ufunc(ufunc, method, inputs, kwargs, self._classes, self._functions)

    def snapshot(self):
        return awkward1._util.wrap(self._fillablearray.snapshot(), self._classes, self._functions)

    def null(self):
        self._fillablearray.null()

    def boolean(self, x):
        self._fillablearray.boolean(x)

    def integer(self, x):
        self._fillablearray.integer(x)

    def real(self, x):
        self._fillablearray.real(x)

    def bytestring(self, x):
        self._fillablearray.bytestring(x)

    def string(self, x):
        self._fillablearray.string(x)

    def beginlist(self):
        self._fillablearray.beginlist()

    def endlist(self):
        self._fillablearray.endlist()

    def begintuple(self):
        self._fillablearray.begintuple()

    def index(self, i):
        self._fillablearray.index(i)

    def endtuple(self):
        self._fillablearray.endtuple()

    def beginrecord(self, name=None):
        self._fillablearray.beginrecord(name)

    def field(self, key):
        self._fillablearray.field(key)

    def endrecord(self):
        self._fillablearray.endrecord()



# BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE

import pytest
import numpy

import awkward1

def test_identity():
    a = numpy.arange(10)
    b = awkward1.layout.NumpyArray(a)
    b.setidentities()
    assert numpy.array(b.identities).tolist() == numpy.arange(10).reshape(-1, 1).tolist()

    assert numpy.array(b[3]) == a[3]
    assert numpy.array(b[3:7].identities).tolist() == numpy.arange(10).reshape(-1, 1)[3:7].tolist()
    assert numpy.array(b[[7, 3, 3, -4]].identities).tolist() == numpy.arange(10).reshape(-1, 1)[[7, 3, 3, -4]].tolist()
    assert numpy.array(b[[True, True, True, False, False, False, True, False, True, False]].identities).tolist() == numpy.arange(10).reshape(-1, 1)[[True, True, True, False, False, False, True, False, True, False]].tolist()

    assert numpy.array(b[1:][3]) == a[1:][3]
    assert numpy.array(b[1:][3:7].identities).tolist() == numpy.arange(10).reshape(-1, 1)[1:][3:7].tolist()
    assert numpy.array(b[1:][[7, 3, 3, -4]].identities).tolist() == numpy.arange(10).reshape(-1, 1)[1:][[7, 3, 3, -4]].tolist()
    assert numpy.array(b[1:][[True, True, False, False, False, True, False, True, False]].identities).tolist() == numpy.arange(10).reshape(-1, 1)[1:][[True, True, False, False, False, True, False, True, False]].tolist()



# BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE

import sys

import numpy
import numba
import llvmlite.ir.types
import llvmlite.llvmpy.core

from .._numba import cpu

py27 = (sys.version_info[0] < 3)

if not py27:
    exec("""
def debug(context, builder, *args):
    assert len(args) % 2 == 0
    tpes, vals = args[0::2], args[1::2]
    context.get_function(print, numba.none(*tpes))(builder, tuple(vals))
""", globals())

dynamic_addrs = {}
def globalstring(context, builder, pyvalue):
    if pyvalue not in dynamic_addrs:
        buf = dynamic_addrs[pyvalue] = numpy.array(pyvalue.encode("utf-8") + b"\x00")
        context.add_dynamic_addr(builder, buf.ctypes.data, info="str({0})".format(repr(pyvalue)))
    ptr = context.get_constant(numba.types.uintp, dynamic_addrs[pyvalue].ctypes.data)
    return builder.inttoptr(ptr, llvmlite.llvmpy.core.Type.pointer(llvmlite.llvmpy.core.Type.int(8)))

RefType = numba.int64

index8tpe = numba.types.Array(numba.int8, 1, "C")
indexU8tpe = numba.types.Array(numba.uint8, 1, "C")
index32tpe = numba.types.Array(numba.int32, 1, "C")
indexU32tpe = numba.types.Array(numba.uint32, 1, "C")
index64tpe = numba.types.Array(numba.int64, 1, "C")
def indextpe(indexname):
    if indexname == "64":
        return index64tpe
    elif indexname == "32":
        return index32tpe
    elif indexname == "U32":
        return indexU32tpe
    elif indexname == "8":
        return index8tpe
    elif indexname == "U8":
        return indexU8tpe
    else:
        raise AssertionError("unrecognized index type: {0}".format(indexname))

def cast(context, builder, fromtpe, totpe, val):
    if isinstance(fromtpe, llvmlite.ir.types.IntType):
        if fromtpe.width == 8:
            fromtpe = numba.int8
        elif fromtpe.width == 16:
            fromtpe = numba.int16
        elif fromtpe.width == 32:
            fromtpe = numba.int32
        elif fromtpe.width == 64:
            fromtpe = numba.int64
        else:
            raise AssertionError("unrecognized bitwidth")
    if fromtpe.bitwidth < totpe.bitwidth:
        return builder.sext(val, context.get_value_type(totpe))
    elif fromtpe.bitwidth > totpe.bitwidth:
        return builder.trunc(val, context.get_value_type(totpe))
    else:
        return val

def arrayptr(context, builder, tpe, val):
    return numba.targets.arrayobj.make_array(tpe)(context, builder, val).data

def arraylen(context, builder, tpe, val, totpe=None):
    if isinstance(tpe, numba.types.Array):
        out = numba.targets.arrayobj.array_len(context, builder, numba.intp(tpe), (val,))
    else:
        out = tpe.lower_len(context, builder, numba.intp(tpe), (val,))
    if totpe is None:
        return out
    else:
        return cast(context, builder, numba.intp, totpe, out)

def call(context, builder, fcn, args, errormessage=None):
    fcntpe = context.get_function_pointer_type(fcn.numbatpe)
    fcnval = context.add_dynamic_addr(builder, fcn.numbatpe.get_pointer(fcn), info=fcn.name)
    fcnptr = builder.bitcast(fcnval, fcntpe)

    err = context.call_function_pointer(builder, fcnptr, args)

    if fcn.restype is cpu.Error:
        assert errormessage is not None, "this function can return an error"
        proxyerr = numba.cgutils.create_struct_proxy(cpu.Error.numbatpe)(context, builder, value=err)
        with builder.if_then(builder.icmp_signed("!=", proxyerr.str, context.get_constant(numba.intp, 0)), likely=False):
            context.call_conv.return_user_exc(builder, ValueError, (errormessage,))

def newindex8(context, builder, lentpe, lenval):
    return numba.targets.arrayobj.numpy_empty_nd(context, builder, index8tpe(lentpe), (lenval,))
def newindexU8(context, builder, lentpe, lenval):
    return numba.targets.arrayobj.numpy_empty_nd(context, builder, indexU8tpe(lentpe), (lenval,))
def newindex32(context, builder, lentpe, lenval):
    return numba.targets.arrayobj.numpy_empty_nd(context, builder, index32tpe(lentpe), (lenval,))
def newindexU32(context, builder, lentpe, lenval):
    return numba.targets.arrayobj.numpy_empty_nd(context, builder, indexU32tpe(lentpe), (lenval,))
def newindex64(context, builder, lentpe, lenval):
    return numba.targets.arrayobj.numpy_empty_nd(context, builder, index64tpe(lentpe), (lenval,))
def newindex(indexname, context, builder, lentpe, lenval):
    if indexname == "64":
        return newindex64(context, builder, lentpe, lenval)
    elif indexname == "32":
        return newindex32(context, builder, lentpe, lenval)
    elif indexname == "U32":
        return newindexU32(context, builder, lentpe, lenval)
    elif indexname == "8":
        return newindex8(context, builder, lentpe, lenval)
    elif indexname == "U8":
        return newindexU8(context, builder, lentpe, lenval)
    else:
        raise AssertionError("unrecognized index type: {0}".format(indexname))

@numba.jit(nopython=True)
def shapeat(shapeat, array, at, ndim):
    redat = at - (ndim - array.ndim)
    if redat < 0:
        return 1
    elif shapeat == 0:
        return 0
    elif shapeat == 1:
        return array.shape[redat]
    elif shapeat == array.shape[redat] or array.shape[redat] == 1:
        return shapeat
    else:
        raise ValueError("cannot broadcast arrays to the same shape")

@numba.generated_jit(nopython=True)
def broadcast_to(array, shape):
    if isinstance(array, numba.types.Array):
        def impl(array, shape):
            out = numpy.empty(shape, array.dtype)
            out[:] = array
            return out
        return impl
    elif isinstance(array, numba.types.Integer):
        def impl(array, shape):
            return numpy.full(shape, array, numpy.int64)
        return impl
    else:
        return lambda array, shape: array

@numba.generated_jit(nopython=True)
def broadcast_arrays(arrays):
    if not isinstance(arrays, numba.types.BaseTuple) or not any(isinstance(x, numba.types.Array) for x in arrays.types):
        return lambda arrays: arrays

    else:
        ndim = max(t.ndim if isinstance(t, numba.types.Array) else 1 for t in arrays.types)
        def getshape(i, at):
            if i == -1:
                return "1"
            elif isinstance(arrays.types[i], numba.types.Array):
                return "shapeat({0}, arrays[{1}], {2}, {3})".format(getshape(i - 1, at), i, at, ndim)
            else:
                return getshape(i - 1, at)
        g = {"shapeat": shapeat, "broadcast_to": broadcast_to}
        exec("""
def impl(arrays):
    shape = ({0})
    return ({1})
""".format(" ".join(getshape(len(arrays.types) - 1, at) + "," for at in range(ndim)),
           " ".join("broadcast_to(arrays[{0}], shape),".format(at) if isinstance(arrays.types[at], (numba.types.Array, numba.types.Integer)) else "arrays[{0}],".format(at) for at in range(len(arrays.types)))), g)
        return g["impl"]

def typing_broadcast_arrays(arrays):
    if not isinstance(arrays, numba.types.BaseTuple) or not any(isinstance(x, numba.types.Array) for x in arrays.types):
        return arrays
    else:
        return numba.types.Tuple([numba.types.Array(numba.int64, 1, "C") if isinstance(t, numba.types.Integer) else t for t in arrays.types])

@numba.generated_jit(nopython=True)
def regularize_slice(arrays):
    if not isinstance(arrays, numba.types.BaseTuple) and isinstance(arrays, (numba.types.ArrayCompatible, numba.types.List)) and isinstance(arrays.dtype, numba.types.Boolean):
        return lambda arrays: numpy.nonzero(arrays)

    elif not isinstance(arrays, numba.types.BaseTuple) or not any(isinstance(t, (numba.types.ArrayCompatible, numba.types.List)) for t in arrays.types):
        return lambda arrays: arrays

    else:
        code = "def impl(arrays):\n"
        indexes = []   # all entries have trailing commas to ensure output is a tuple
        for i, t in enumerate(arrays.types):
            if isinstance(t, (numba.types.ArrayCompatible, numba.types.List)) and isinstance(t.dtype, numba.types.Boolean):
                code += "    x{0} = numpy.nonzero(arrays[{1}])\n".format(i, i)
                indexes.extend(["x{0}[{1}],".format(i, j) for j in range(arrays.types[i].ndim)])
            elif isinstance(t, (numba.types.ArrayCompatible, numba.types.List)) and isinstance(t.dtype, numba.types.Integer):
                indexes.append("numpy.asarray(arrays[{0}], numpy.int64),".format(i))
            elif isinstance(t, (numba.types.ArrayCompatible, numba.types.List)):
                raise TypeError("arrays must have boolean or integer type")
            else:
                indexes.append("arrays[{0}],".format(i))
        code += "    return ({0})".format(" ".join(indexes))
        g = {"numpy": numpy}
        exec(code, g)
        return g["impl"]

def typing_regularize_slice(arrays):
    out = ()
    if not isinstance(arrays, numba.types.BaseTuple) and isinstance(arrays, (numba.types.ArrayCompatible, numba.types.List)) and isinstance(arrays.dtype, numba.types.Boolean):
        return numba.types.Tuple(arrays.ndims*(numba.types.Array(numba.int64, 1, "C"),))

    elif not isinstance(arrays, numba.types.BaseTuple) or not any(isinstance(t, (numba.types.ArrayCompatible, numba.types.List)) for t in arrays.types):
        return arrays

    else:
        for t in arrays.types:
            if isinstance(t, (numba.types.ArrayCompatible, numba.types.List)) and isinstance(t.dtype, numba.types.Boolean):
                out = out + t.ndims*(numba.types.Array(numba.int64, 1, "C"),)
            elif isinstance(t, (numba.types.ArrayCompatible, numba.types.List)) and isinstance(t.dtype, numba.types.Integer):
                out = out + (numba.types.Array(numba.int64, 1, "C"),)
            elif isinstance(t, (numba.types.ArrayCompatible, numba.types.List)):
                raise TypeError("arrays must have boolean or integer type")
            else:
                out = out + (t,)
        return numba.types.Tuple(out)

def preprocess_slicetuple(context, builder, wheretpe, whereval):
    wheretpe2 = typing_regularize_slice(wheretpe)
    regularize_slice.compile(wheretpe2(wheretpe))
    cres = regularize_slice.overloads[(wheretpe,)]
    whereval2 = context.call_internal(builder, cres.fndesc, wheretpe2(wheretpe), (whereval,))

    wheretpe3 = typing_broadcast_arrays(wheretpe2)
    broadcast_arrays.compile(wheretpe3(wheretpe2))
    cres2 = broadcast_arrays.overloads[(wheretpe2,)]
    whereval3 = context.call_internal(builder, cres2.fndesc, wheretpe3(wheretpe2), (whereval2,))

    return wheretpe3, whereval3

def preserves_type(slice, isadvanced):
    if isinstance(slice, numba.types.Integer):
        return False
    elif isinstance(slice, numba.types.SliceType):
        return True
    elif isinstance(slice, numba.types.EllipsisType):
        return True
    elif isinstance(slice, type(numba.typeof(numpy.newaxis))):
        return False
    elif isinstance(slice, (numba.types.Array, numba.types.List, numba.types.ArrayCompatible)):
        return not isadvanced
    elif isinstance(slice, numba.types.StringLiteral):
        return False
    else:
        raise AssertionError(slice)

def dict2parameters(parameters):
    return tuple(sorted(parameters.items()))

def parameters2str(parameters):
    return "{" + ", ".join("{0}: {1}".format(n, x) for n, x in parameters) + "}"

def parameters2dict_impl(c, parameters):
    dict_obj = c.pyapi.unserialize(c.pyapi.serialize_object(dict))
    parameters_obj = c.pyapi.unserialize(c.pyapi.serialize_object(parameters))
    out = c.pyapi.call_function_objargs(dict_obj, (parameters_obj,))
    c.pyapi.decref(dict_obj)
    c.pyapi.decref(parameters_obj)
    return out



# BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE

import sys
import itertools

import pytest
import numpy

numba = pytest.importorskip("numba")

import awkward1

content1 = awkward1.layout.NumpyArray(numpy.array([1, 2, 3, 4, 5]))
content2 = awkward1.layout.NumpyArray(numpy.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9]))
offsets = awkward1.layout.Index64(numpy.array([0, 3, 3, 5, 6, 9]))
listoffsetarray = awkward1.layout.ListOffsetArray64(offsets, content2)
recordarray = awkward1.layout.RecordArray({"one": content1, "two": listoffsetarray})

def test_boxing():
    @numba.njit
    def f1(q):
        return 3.14

    assert f1(recordarray) == 3.14
    assert f1(recordarray[2]) == 3.14

    @numba.njit
    def f2(q):
        return q

    assert awkward1.tolist(f2(recordarray)) == [{"one": 1, "two": [1.1, 2.2, 3.3]}, {"one": 2, "two": []}, {"one": 3, "two": [4.4, 5.5]}, {"one": 4, "two": [6.6]}, {"one": 5, "two": [7.7, 8.8, 9.9]}]
    assert awkward1.tolist(f2(recordarray[2])) == {"one": 3, "two": [4.4, 5.5]}

def test_len():
    @numba.njit
    def f1(q):
        return len(q)

    assert f1(recordarray) == 5
    with pytest.raises(numba.TypingError):
        f1(recordarray[2])

def test_getitem_int():
    @numba.njit
    def f1(q):
        return q[2]

    assert awkward1.tolist(f1(recordarray)) == {"one": 3, "two": [4.4, 5.5]}
    with pytest.raises(numba.TypingError):
        f1(recordarray[2])

def test_getitem_iter():
    @numba.njit
    def f1(q):
        out = 0
        for x in q:
            out += 1
        return out

    assert f1(recordarray) == 5
    with pytest.raises(numba.TypingError):
        f1(recordarray[2])

def test_getitem_range():
    @numba.njit
    def f1(q):
        return q[1:4]

    assert awkward1.tolist(f1(recordarray)) == [{"one": 2, "two": []}, {"one": 3, "two": [4.4, 5.5]}, {"one": 4, "two": [6.6]}]
    with pytest.raises(numba.TypingError):
        f1(recordarray[2])

def test_getitem_str():
    outer_starts = numpy.array([0, 3, 3], dtype=numpy.int64)
    outer_stops = numpy.array([3, 3, 5], dtype=numpy.int64)
    outer_offsets = numpy.array([0, 3, 3, 5], dtype=numpy.int64)
    outer_listarray = awkward1.layout.ListArray64(awkward1.layout.Index64(outer_starts), awkward1.layout.Index64(outer_stops), recordarray)
    outer_listoffsetarray = awkward1.layout.ListOffsetArray64(awkward1.layout.Index64(outer_offsets), recordarray)
    outer_regulararray = awkward1.layout.RegularArray(recordarray, 2)

    @numba.njit
    def f1(q):
        return q["one"]

    assert awkward1.tolist(f1(recordarray)) == [1, 2, 3, 4, 5]

    assert sys.getrefcount(outer_starts), sys.getrefcount(outer_stops) == (3, 3)
    assert awkward1.tolist(f1(outer_listarray)) == [[1, 2, 3], [], [4, 5]]
    assert sys.getrefcount(outer_starts), sys.getrefcount(outer_stops) == (3, 3)

    assert sys.getrefcount(outer_offsets) == 3
    assert awkward1.tolist(f1(outer_listoffsetarray)) == [[1, 2, 3], [], [4, 5]]
    assert sys.getrefcount(outer_offsets) == 3

    assert awkward1.tolist(f1(outer_regulararray)) == [[1, 2], [3, 4]]

    @numba.njit
    def f2(q):
        return q["two"]

    assert awkward1.tolist(f1(recordarray[2])) == 3
    assert awkward1.tolist(f2(recordarray[2])) == [4.4, 5.5]

def test_getitem_tuple():
    content3 = awkward1.layout.NumpyArray(numpy.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10]))
    regulararray = awkward1.layout.RegularArray(content3, 2)
    offsets2 = awkward1.layout.Index64(numpy.array([0, 3, 4, 5, 8, 9]))
    listoffsetarray2 = awkward1.layout.ListOffsetArray64(offsets2, content2)
    recordarray2 = awkward1.layout.RecordArray({"one": regulararray, "two": listoffsetarray2})

    assert awkward1.tolist(recordarray2) == [{"one": [1, 2], "two": [1.1, 2.2, 3.3]}, {"one": [3, 4], "two": [4.4]}, {"one": [5, 6], "two": [5.5]}, {"one": [7, 8], "two": [6.6, 7.7, 8.8]}, {"one": [9, 10], "two": [9.9]}]

    @numba.njit
    def f1(q):
        return q[:, -1]

    assert awkward1.tolist(f1(recordarray2)) == [{"one": 2, "two": 3.3}, {"one": 4, "two": 4.4}, {"one": 6, "two": 5.5}, {"one": 8, "two": 8.8}, {"one": 10, "two": 9.9}]

    @numba.njit
    def f2(q):
        return q[:, -2:]

    assert awkward1.tolist(f2(recordarray2)) == [{"one": [1, 2], "two": [2.2, 3.3]}, {"one": [3, 4], "two": [4.4]}, {"one": [5, 6], "two": [5.5]}, {"one": [7, 8], "two": [7.7, 8.8]}, {"one": [9, 10], "two": [9.9]}]

    @numba.njit
    def f3(q):
        return q[2:, "two"]

    assert awkward1.tolist(f3(recordarray2)) == [[5.5], [6.6, 7.7, 8.8], [9.9]]

    @numba.njit
    def f4(q):
        return q["two", 1:-1]

    assert awkward1.tolist(f4(recordarray2)) == [[4.4], [5.5], [6.6, 7.7, 8.8]]

def test_fillablearray_tuple():
    fillablearray = awkward1.layout.FillableArray()

    @numba.njit
    def f1(q):
        q.begintuple(3)
        q.index(0); q.boolean(True)
        q.index(1); q.integer(1)
        q.index(2); q.real(1.1)
        q.endtuple()

        q.begintuple(3)
        q.index(0); q.boolean(False)
        q.index(1); q.integer(1)
        q.index(2); q.real(1.1)
        q.endtuple()

        q.begintuple(3)
        q.index(0); q.boolean(True)
        q.index(1); q.integer(1)
        q.index(2); q.real(1.1)
        q.endtuple()

        return q

    fillablearray2 = f1(fillablearray)

    assert awkward1.tolist(fillablearray.snapshot()) == [(True, 1, 1.1), (False, 1, 1.1), (True, 1, 1.1)]
    assert awkward1.tolist(fillablearray2.snapshot()) == [(True, 1, 1.1), (False, 1, 1.1), (True, 1, 1.1)]

def test_fillablearray_record_1():
    fillablearray = awkward1.layout.FillableArray()

    @numba.njit
    def f1(q):
        q.beginrecord()
        q.field("one"); q.boolean(True)
        q.field("two"); q.integer(1)
        q.field("three"); q.real(1.1)
        q.endrecord()

        q.beginrecord()
        q.field("one"); q.boolean(False)
        q.field("two"); q.integer(2)
        q.field("three"); q.real(2.2)
        q.endrecord()

        q.beginrecord()
        q.field("one"); q.boolean(True)
        q.field("two"); q.integer(3)
        q.field("three"); q.real(3.3)
        q.endrecord()

        return q

    fillablearray2 = f1(fillablearray)

    assert awkward1.tolist(fillablearray.snapshot()) == [{'one': True, 'two': 1, 'three': 1.1}, {'one': False, 'two': 2, 'three': 2.2}, {'one': True, 'two': 3, 'three': 3.3}]
    assert awkward1.tolist(fillablearray2.snapshot()) == [{'one': True, 'two': 1, 'three': 1.1}, {'one': False, 'two': 2, 'three': 2.2}, {'one': True, 'two': 3, 'three': 3.3}]

def test_fillablearray_record_2():
    fillablearray = awkward1.layout.FillableArray()

    @numba.njit
    def f1(q):
        q.beginrecord("wowie")
        q.field("one"); q.boolean(True)
        q.field("two"); q.integer(1)
        q.field("three"); q.real(1.1)
        q.endrecord()

        q.beginrecord("wowie")
        q.field("one"); q.boolean(False)
        q.field("two"); q.integer(2)
        q.field("three"); q.real(2.2)
        q.endrecord()

        q.beginrecord("wowie")
        q.field("one"); q.boolean(True)
        q.field("two"); q.integer(3)
        q.field("three"); q.real(3.3)
        q.endrecord()

        return q

    fillablearray2 = f1(fillablearray)

    assert awkward1.tolist(fillablearray.snapshot()) == [{'one': True, 'two': 1, 'three': 1.1}, {'one': False, 'two': 2, 'three': 2.2}, {'one': True, 'two': 3, 'three': 3.3}]
    assert awkward1.tolist(fillablearray2.snapshot()) == [{'one': True, 'two': 1, 'three': 1.1}, {'one': False, 'two': 2, 'three': 2.2}, {'one': True, 'two': 3, 'three': 3.3}]



# BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE

import numba

import awkward1.layout
from .._numba import cpu, util, content

class IteratorType(numba.types.common.SimpleIteratorType):
    def __init__(self, arraytpe):
        self.arraytpe = arraytpe
        super(IteratorType, self).__init__("ak::Iterator({0})".format(self.arraytpe.name), self.arraytpe.getitem_int())

@numba.typing.templates.infer
class ContentType_type_getiter(numba.typing.templates.AbstractTemplate):
    key = "getiter"

    def generic(self, args, kwargs):
        if len(args) == 1 and len(kwargs) == 0:
            arraytpe, = args
            if isinstance(arraytpe, content.ContentType):
                return IteratorType(arraytpe)(arraytpe)

@numba.datamodel.registry.register_default(IteratorType)
class IteratorModel(numba.datamodel.models.StructModel):
    def __init__(self, dmm, fe_type):
        members = [("array", fe_type.arraytpe),
                   ("length", numba.int64),
                   ("at", numba.types.EphemeralPointer(numba.int64))]
        super(IteratorModel, self).__init__(dmm, fe_type, members)

@numba.extending.lower_builtin("getiter", content.ContentType)
def lower_getiter(context, builder, sig, args):
    rettpe, (tpe,) = sig.return_type, sig.args
    val, = args
    proxyout = context.make_helper(builder, rettpe)
    proxyout.array = val
    proxyout.length = util.cast(context, builder, numba.intp, numba.int64, tpe.lower_len(context, builder, numba.intp(tpe), (val,)))
    proxyout.at = numba.cgutils.alloca_once_value(builder, context.get_constant(numba.int64, 0))
    if context.enable_nrt:
        context.nrt.incref(builder, tpe, val)
    return numba.targets.imputils.impl_ret_new_ref(context, builder, rettpe, proxyout._getvalue())

@numba.extending.lower_builtin("iternext", IteratorType)
@numba.targets.imputils.iternext_impl(numba.targets.imputils.RefType.BORROWED)
def lower_iternext(context, builder, sig, args, result):
    tpe, = sig.args
    val, = args

    proxyin = context.make_helper(builder, tpe, value=val)
    at = builder.load(proxyin.at)

    is_valid = builder.icmp_signed("<", at, proxyin.length)
    result.set_valid(is_valid)

    with builder.if_then(is_valid, likely=True):
        result.yield_(tpe.arraytpe.lower_getitem_int(context, builder, tpe.yield_type(tpe.arraytpe, numba.int64), (proxyin.array, at)))
        nextat = numba.cgutils.increment_index(builder, at)
        builder.store(nextat, proxyin.at)



# BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE

import sys

import pytest
import numpy

import awkward1

def flatten(data, axis=0):
    if axis < 0:
        raise NotImplementedError("axis < 0 is much harder for untyped data...")
    if isinstance(data, list):
        if axis == 0:
            if all(isinstance(x, list) for x in data):
                return sum(data, [])
            else:
                raise ValueError("cannot concatenate non-lists")
        else:
            return [flatten(x, axis - 1) for x in data]
    else:
        raise ValueError("cannot flatten {0} objects".format(type(data)))

def test_flatten_empty_array():
    empty = awkward1.layout.EmptyArray()

    assert awkward1.tolist(empty) == []
    assert awkward1.tolist(empty.flatten()) == []

def test_flatten_list_array():
    content = awkward1.layout.NumpyArray(numpy.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9]))
    starts  = awkward1.layout.Index64(numpy.array([0, 3, 4, 5, 8]))
    stops   = awkward1.layout.Index64(numpy.array([3, 3, 6, 8, 9]))
    array   = awkward1.layout.ListArray64(starts, stops, content)

    assert awkward1.tolist(array) == [[0.0, 1.1, 2.2], [], [4.4, 5.5], [5.5, 6.6, 7.7], [8.8]]
    assert awkward1.tolist(array.flatten()) == [0.0, 1.1, 2.2, 4.4, 5.5, 5.5, 6.6, 7.7, 8.8]

    array2 = array[2:-1]
    assert awkward1.tolist(array2) == [[4.4, 5.5], [5.5, 6.6, 7.7]]
    assert flatten(awkward1.tolist(array2)) == [4.4, 5.5, 5.5, 6.6, 7.7]
    assert awkward1.tolist(array2.flatten()) == [4.4, 5.5, 5.5, 6.6, 7.7]

    # The following are allowed:
    #     * out of order (4:7 before 0:1)
    #     * overlaps (0:1 and 0:4 and 1:5)
    #     * beyond content starts[i] == stops[i] (999)
    #
    # See https://github.com/scikit-hep/awkward-1.0/wiki/ListArray.md
    #
    content = awkward1.layout.NumpyArray(numpy.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9], dtype=numpy.int64))
    starts  = awkward1.layout.Index64(numpy.array([4, 999, 0, 0, 1, 7]))
    stops   = awkward1.layout.Index64(numpy.array([7, 999, 1, 4, 5, 10]))
    array   = awkward1.layout.ListArray64(starts, stops, content)
    assert awkward1.tolist(array) == [[4, 5, 6], [], [0], [0, 1, 2, 3], [1, 2, 3, 4], [7, 8, 9]]
    assert flatten(awkward1.tolist(array)) == [4, 5, 6, 0, 0, 1, 2, 3, 1, 2, 3, 4, 7, 8, 9]
    assert awkward1.tolist(array.flatten()) == [4, 5, 6, 0, 0, 1, 2, 3, 1, 2, 3, 4, 7, 8, 9]

def test_flatten_list_offset_array():
    content = awkward1.layout.NumpyArray(numpy.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9]))
    offsets = awkward1.layout.Index64(numpy.array([0, 3, 3, 5, 6, 10], dtype=numpy.int64))
    array   = awkward1.layout.ListOffsetArray64(offsets, content)

    assert awkward1.tolist(array) == [[0.0, 1.1, 2.2], [], [3.3, 4.4], [5.5], [6.6, 7.7, 8.8, 9.9]]
    assert awkward1.tolist(array.flatten()) == [0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9]

    array2 = array[2:-1]
    assert awkward1.tolist(array2.flatten()) == [3.3, 4.4, 5.5]

def test_flatten_numpy_array():
    array = awkward1.layout.NumpyArray(numpy.arange(2*3*5, dtype=numpy.int64).reshape(2, 3, 5))
    assert awkward1.tolist(array) == [[[ 0,  1,  2,  3,  4], [ 5,  6,  7,  8,  9], [10, 11, 12, 13, 14]],
                                      [[15, 16, 17, 18, 19], [20, 21, 22, 23, 24], [25, 26, 27, 28, 29]]]
    assert awkward1.tolist(array.flatten()) == [[0, 1, 2, 3, 4], [5, 6, 7, 8, 9], [10, 11, 12, 13, 14], [15, 16, 17, 18, 19], [20, 21, 22, 23, 24], [25, 26, 27, 28, 29]]

    array2 = awkward1.layout.NumpyArray(numpy.arange(1*6*5, dtype=numpy.int64).reshape(1, 6, 5))
    assert awkward1.tolist(array2.flatten()) == [[0, 1, 2, 3, 4], [5, 6, 7, 8, 9], [10, 11, 12, 13, 14], [15, 16, 17, 18, 19], [20, 21, 22, 23, 24], [25, 26, 27, 28, 29]]

    array3 = array[:, 1::2, :3]
    print(awkward1.tolist(array3))
    assert awkward1.tolist(array3) == [[[5, 6, 7]], [[20, 21, 22]]]
    assert flatten(awkward1.tolist(array3)) == [[5, 6, 7], [20, 21, 22]]
    assert awkward1.tolist(array3.flatten()) == [[5, 6, 7], [20, 21, 22]]

## def test_flatten_raw_array():
    ## RawArrayOf<T> is usable only in C++

def test_flatten_record():
    content = awkward1.layout.NumpyArray(numpy.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9]))
    offsets = awkward1.layout.Index64(numpy.array([0, 3, 3, 5, 6, 10], dtype=numpy.int64))
    #FIXME: array   = awkward1.layout.Record()

def test_flatten_record_array():
    with pytest.raises(ValueError, match="NumpyArray cannot be flattened because it has 1 dimensions"):
        content1 = awkward1.layout.NumpyArray(numpy.array([1, 2, 3, 4, 5]))
        content2 = awkward1.layout.NumpyArray(numpy.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9]))
        offsets = awkward1.layout.Index64(numpy.array([0, 3, 3, 5, 6, 9]))
        listoffsetarray = awkward1.layout.ListOffsetArray64(offsets, content2)
        recordarray = awkward1.layout.RecordArray({"one": content1, "two": listoffsetarray})

        assert awkward1.tolist(recordarray) == [{'one': 1, 'two': [1.1, 2.2, 3.3]}, {'one': 2, 'two': []}, {'one': 3, 'two': [4.4, 5.5]}, {'one': 4, 'two': [6.6]}, {'one': 5, 'two': [7.7, 8.8, 9.9]}]
        assert awkward1.tolist(recordarray.flatten()) == [{'one': [1, 2, 3, 4, 5], 'two': [1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9]}]

def test_flatten_regular_array():
    content = awkward1.layout.NumpyArray(numpy.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9]))
    offsets = awkward1.layout.Index64(numpy.array([0, 3, 3, 5, 6, 10, 10]))
    listoffsetarray = awkward1.layout.ListOffsetArray64(offsets, content)
    regulararray = awkward1.layout.RegularArray(listoffsetarray, 2)

    assert awkward1.tolist(regulararray) == [[[0.0, 1.1, 2.2], []], [[3.3, 4.4], [5.5]], [[6.6, 7.7, 8.8, 9.9], []]]
    assert awkward1.tolist(regulararray.flatten()) == [[0.0, 1.1, 2.2], [], [3.3, 4.4], [5.5], [6.6, 7.7, 8.8, 9.9], []]

    ## From https://github.com/scikit-hep/awkward-array example:
    ## array:  [[[1.1 2.2] [3.3]] [] [[4.4 5.5]] [[6.6 7.7 8.8] [] [9.9]]]
    ## axis 0: [[1.1 2.2] [3.3] [4.4 5.5] [6.6 7.7 8.8] [] [9.9]]
    ## axis 1: [[1.1 2.2 3.3] [] [4.4 5.5] [6.6 7.7 8.8 9.9]]

    ## array:  [[[0.0, 1.1, 2.2], []], [[3.3, 4.4], [5.5]], [[6.6, 7.7, 8.8, 9.9], []]]
    ## axis 0: [[0.0, 1.1, 2.2], [], [3.3, 4.4], [5.5], [6.6, 7.7, 8.8, 9.9], []]
    ## axis 1: [[0.0, 1.1, 2.2], [3.3, 4.4, 5.5], [6.6, 7.7, 8.8, 9.9]]

    regulararray_m1 = awkward1.layout.RegularArray(listoffsetarray[:-1], 2)

    assert awkward1.tolist(regulararray_m1) == [[[0.0, 1.1, 2.2], []], [[3.3, 4.4], [5.5]]]
    assert awkward1.tolist(regulararray_m1.flatten()) == [[0.0, 1.1, 2.2], [], [3.3, 4.4], [5.5]]

    regulararray_m2 = awkward1.layout.RegularArray(listoffsetarray[:-2], 2)

    assert awkward1.tolist(regulararray_m2) == [[[0.0, 1.1, 2.2], []], [[3.3, 4.4], [5.5]]]
    assert awkward1.tolist(regulararray_m2.flatten()) == [[0.0, 1.1, 2.2], [], [3.3, 4.4], [5.5]]



# BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE

import inspect
import numbers
import re

import numpy

import awkward1.layout

unknowntypes = (awkward1.layout.EmptyArray,)

indexedtypes = (awkward1.layout.IndexedArray32, awkward1.layout.IndexedArrayU32, awkward1.layout.IndexedArray64)

uniontypes = (awkward1.layout.UnionArray8_32, awkward1.layout.UnionArray8_U32, awkward1.layout.UnionArray8_64)

optiontypes = (awkward1.layout.IndexedOptionArray32, awkward1.layout.IndexedOptionArray64)

listtypes = (awkward1.layout.RegularArray, awkward1.layout.ListArray32, awkward1.layout.ListArrayU32, awkward1.layout.ListArray64, awkward1.layout.ListOffsetArray32, awkward1.layout.ListOffsetArrayU32, awkward1.layout.ListOffsetArray64)

recordtypes = (awkward1.layout.RecordArray,)

def regular_classes(classes):
    import awkward1
    if classes is None:
        return awkward1.classes
    else:
        return classes

def regular_functions(functions):
    import awkward1
    if functions is None:
        return awkward1.functions
    else:
        return functions

def wrap(content, classes, functions):
    import awkward1.layout

    if isinstance(content, awkward1.layout.Content):
        cls = regular_classes(classes).get(content.parameters.get("__class__"))
        if cls is None or (isinstance(cls, type) and not issubclass(cls, awkward1.Array)):
            cls = awkward1.Array
        return cls(content, classes=classes, functions=functions)

    elif isinstance(content, awkward1.layout.Record):
        cls = regular_classes(classes).get(content.parameters.get("__class__"))
        if cls is None or (isinstance(cls, type) and not issubclass(cls, awkward1.Record)):
            cls = awkward1.Record
        return cls(content, classes=classes, functions=functions)

    else:
        return content

def called_by_module(modulename):
    frame = inspect.currentframe()
    while frame is not None:
        name = inspect.getmodule(frame).__name__
        if name == modulename or name.startswith(modulename + "."):
            return True
        frame = frame.f_back
    return True

def key2index(keys, key):
    if keys is None:
        attempt = None
    else:
        try:
            attempt = keys.index(key)
        except ValueError:
            attempt = None

    if attempt is None:
        m = key2index._pattern.match(key)
        if m is not None:
            attempt = m.group(0)

    if attempt is None:
        raise ValueError("key {0} not found in record".format(repr(key)))
    else:
        return attempt

key2index._pattern = re.compile(r"^[1-9][0-9]*$")

def minimally_touching_string(limit_length, layout, classes, functions):
    import awkward1.layout

    if isinstance(layout, awkward1.layout.Record):
        layout = layout.array[layout.at : layout.at + 1]

    if len(layout) == 0:
        return "[]"

    def forward(x, space, brackets=True, wrap=True):
        done = False
        if wrap and isinstance(x, awkward1.layout.Content):
            cls = regular_classes(classes).get(x.parameters.get("__class__"))
            if cls is not None and isinstance(cls, type) and issubclass(cls, awkward1.Array):
                y = cls(x, classes=classes, functions=functions)
                if "__repr__" in type(y).__dict__:
                    yield space + repr(y)
                    done = True
        if wrap and isinstance(x, awkward1.layout.Record):
            cls = regular_classes(classes).get(x.parameters.get("__class__"))
            if cls is not None and isinstance(cls, type) and issubclass(cls, awkward1.Record):
                y = cls(x, classes=classes, functions=functions)
                if "__repr__" in type(y).__dict__:
                    yield space + repr(y)
                    done = True
        if not done:
            if isinstance(x, awkward1.layout.Content):
                if brackets:
                    yield space + "["
                sp = ""
                for i in range(len(x)):
                    for token in forward(x[i], sp):
                        yield token
                    sp = ", "
                if brackets:
                    yield "]"
            elif isinstance(x, awkward1.layout.Record):
                yield space + "{"
                sp = ""
                for k in x.keys():
                    key = sp + k + ": "
                    for token in forward(x[k], ""):
                        yield key + token
                        key = ""
                    sp = ", "
                yield "}"
            elif isinstance(x, (float, numpy.floating)):
                yield space + "{0:.3g}".format(x)
            else:
                yield space + repr(x)

    def backward(x, space, brackets=True, wrap=True):
        done = False
        if wrap and isinstance(x, awkward1.layout.Content):
            cls = regular_classes(classes).get(x.parameters.get("__class__"))
            if cls is not None and isinstance(cls, type) and issubclass(cls, awkward1.Array):
                y = cls(x, classes=classes, functions=functions)
                if "__repr__" in type(y).__dict__:
                    yield repr(y) + space
                    done = True
        if wrap and isinstance(x, awkward1.layout.Record):
            cls = regular_classes(classes).get(x.parameters.get("__class__"))
            if cls is not None and isinstance(cls, type) and issubclass(cls, awkward1.Record):
                y = cls(x, classes=classes, functions=functions)
                if "__repr__" in type(y).__dict__:
                    yield repr(y) + space
                    done = True
        if not done:
            if isinstance(x, awkward1.layout.Content):
                if brackets:
                    yield "]" + space
                sp = ""
                for i in range(len(x) - 1, -1, -1):
                    for token in backward(x[i], sp):
                        yield token
                    sp = ", "
                if brackets:
                    yield "["
            elif isinstance(x, awkward1.layout.Record):
                yield "}" + space
                keys = x.keys()
                for i in range(len(keys) - 1, -1, -1):
                    last = None
                    for token in backward(x[keys[i]], ""):
                        if last is not None:
                            yield last
                        last = token
                    if last is not None:
                        yield keys[i] + ": " + last
                    if i != 0:
                        yield ", "
                yield "{"
            elif isinstance(x, (float, numpy.floating)):
                yield "{0:.3g}".format(x) + space
            else:
                yield repr(x) + space

    def forever(iterable):
        for token in iterable:
            yield token
        while True:
            yield None

    halfway = len(layout) // 2
    left, right = ["["], ["]"]
    leftlen, rightlen = 1, 1
    leftgen = forever(forward(layout[:halfway], "", brackets=False, wrap=False))
    rightgen = forever(backward(layout[halfway:], "", brackets=False, wrap=False))
    while True:
        l = next(leftgen)
        if l is not None:
            if leftlen + rightlen + len(l) + (2 if l is None and r is None else 6) > limit_length:
                break
            left.append(l)
            leftlen += len(l)

        r = next(rightgen)
        if r is not None:
            if leftlen + rightlen + len(r) + (2 if l is None and r is None else 6) > limit_length:
                break
            right.append(r)
            rightlen += len(r)

        if l is None and r is None:
            break

    while len(left) > 1 and (left[-1] == "[" or left[-1] == ", [" or left[-1] == "{" or left[-1] == ", {" or left[-1] == ", "):
        left.pop()
        l = ""
    while len(right) > 1 and (right[-1] == "]" or right[-1] == "], " or right[-1] == "}" or right[-1] == "}, " or right[-1] == ", "):
        right.pop()
        r = ""
    if l is None and r is None:
        if left == ["["]:
            return "[" + "".join(reversed(right)).lstrip(" ")
        else:
            return "".join(left).rstrip(" ") + ", " + "".join(reversed(right)).lstrip(" ")
    else:
        if left == ["["] and right == ["]"]:
            return "[...]"
        elif left == ["["]:
            return "[... " + "".join(reversed(right)).lstrip(" ")
        else:
            return "".join(left).rstrip(" ") + ", ... " + "".join(reversed(right)).lstrip(" ")



# BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE

import sys

import pytest
import numpy

import awkward1

def test_getitem():
    content0 = awkward1.Array([[1.1, 2.2, 3.3], [], [4.4, 5.5]]).layout
    content1 = awkward1.Array(["one", "two", "three", "four", "five"]).layout
    tags = awkward1.layout.Index8(numpy.array([1, 1, 0, 0, 1, 0, 1, 1], dtype=numpy.int8))

    assert numpy.asarray(awkward1.layout.UnionArray8_32.regular_index(tags)).tolist() == [0, 1, 0, 1, 2, 2, 3, 4]
    assert numpy.asarray(awkward1.layout.UnionArray8_32.regular_index(tags)).dtype == numpy.dtype(numpy.int32)
    assert numpy.asarray(awkward1.layout.UnionArray8_U32.regular_index(tags)).tolist() == [0, 1, 0, 1, 2, 2, 3, 4]
    assert numpy.asarray(awkward1.layout.UnionArray8_U32.regular_index(tags)).dtype == numpy.dtype(numpy.uint32)
    assert numpy.asarray(awkward1.layout.UnionArray8_64.regular_index(tags)).tolist() == [0, 1, 0, 1, 2, 2, 3, 4]
    assert numpy.asarray(awkward1.layout.UnionArray8_64.regular_index(tags)).dtype == numpy.dtype(numpy.int64)

    index = awkward1.layout.Index32(numpy.array([0, 1, 0, 1, 2, 2, 4, 3], dtype=numpy.int32))
    array = awkward1.layout.UnionArray8_32(tags, index, [content0, content1])
    assert numpy.asarray(array.tags).tolist() == [1, 1, 0, 0, 1, 0, 1, 1]
    assert numpy.asarray(array.tags).dtype == numpy.dtype(numpy.int8)
    assert numpy.asarray(array.index).tolist() == [0, 1, 0, 1, 2, 2, 4, 3]
    assert numpy.asarray(array.index).dtype == numpy.dtype(numpy.int32)
    assert type(array.contents) is list
    assert [awkward1.tolist(x) for x in array.contents] == [[[1.1, 2.2, 3.3], [], [4.4, 5.5]], ["one", "two", "three", "four", "five"]]
    assert array.numcontents == 2
    assert awkward1.tolist(array.content(0)) == [[1.1, 2.2, 3.3], [], [4.4, 5.5]]
    assert awkward1.tolist(array.content(1)) == ["one", "two", "three", "four", "five"]
    assert awkward1.tolist(array.project(0)) == [[1.1, 2.2, 3.3], [], [4.4, 5.5]]
    assert awkward1.tolist(array.project(1)) == ["one", "two", "three", "five", "four"]
    repr(array)

    assert awkward1.tolist(array[0]) == "one"
    assert awkward1.tolist(array[1]) == "two"
    assert awkward1.tolist(array[2]) == [1.1, 2.2, 3.3]
    assert awkward1.tolist(array[3]) == []
    assert awkward1.tolist(array[4]) == "three"
    assert awkward1.tolist(array[5]) == [4.4, 5.5]
    assert awkward1.tolist(array[6]) == "five"
    assert awkward1.tolist(array[7]) == "four"

    assert awkward1.tolist(array) == ["one", "two", [1.1, 2.2, 3.3], [], "three", [4.4, 5.5], "five", "four"]
    assert awkward1.tolist(array[1:-1]) == ["two", [1.1, 2.2, 3.3], [], "three", [4.4, 5.5], "five"]
    assert awkward1.tolist(array[2:-2]) == [[1.1, 2.2, 3.3], [], "three", [4.4, 5.5]]
    assert awkward1.tolist(array[::2]) == ["one", [1.1, 2.2, 3.3], "three", "five"]
    assert awkward1.tolist(array[::2, 1:]) == ["ne", [2.2, 3.3], "hree", "ive"]
    assert awkward1.tolist(array[:, :-1]) == ["on", "tw", [1.1, 2.2], [], "thre", [4.4], "fiv", "fou"]

    content2 = awkward1.Array([{"x": 0, "y": []}, {"x": 1, "y": [1.1]}, {"x": 2, "y": [1.1, 2.2]}]).layout
    content3 = awkward1.Array([{"x": 0.0, "y": "zero", "z": False}, {"x": 1.1, "y": "one", "z": True}, {"x": 2.2, "y": "two", "z": False}, {"x": 3.3, "y": "three", "z": True}, {"x": 4.4, "y": "four", "z": False}]).layout
    array2 = awkward1.layout.UnionArray8_32(tags, index, [content2, content3])
    assert awkward1.tolist(array2) == [{"x": 0.0, "y": "zero", "z": False}, {"x": 1.1, "y": "one", "z": True}, {"x": 0, "y": []}, {"x": 1, "y": [1.1]}, {"x": 2.2, "y": "two", "z": False}, {"x": 2, "y": [1.1, 2.2]}, {"x": 4.4, "y": "four", "z": False}, {"x": 3.3, "y": "three", "z": True}]
    assert awkward1.tolist(array2["x"]) == [0.0, 1.1, 0, 1, 2.2, 2, 4.4, 3.3]
    assert awkward1.tolist(array2["y"]) == ["zero", "one", [], [1.1], "two", [1.1, 2.2], "four", "three"]
    assert awkward1.tolist(array2[:, "y", 1:]) == ["ero", "ne", [], [], "wo", [2.2], "our", "hree"]
    assert awkward1.tolist(array2["y", :, 1:]) == ["ero", "ne", [], [], "wo", [2.2], "our", "hree"]
    with pytest.raises(ValueError) as err:
        array2[:, 1:, "y"]
    assert str(err.value) == "in NumpyArray, too many dimensions in slice"
    with pytest.raises(ValueError) as err:
        array2["z"]
    assert str(err.value) == "key \"z\" does not exist (not in record)"

    array3 = awkward1.layout.UnionArray8_32(tags, index, [content3, content2])
    array4 = awkward1.layout.UnionArray8_32(tags, index, [content0, content1, content2, content3])
    assert set(content2.keys()) == set(["x", "y"])
    assert set(content3.keys()) == set(["x", "y", "z"])
    assert set(array2.keys()) == set(["x", "y"])
    assert set(array3.keys()) == set(["x", "y"])
    assert array4.keys() == []

def test_identities():
    content0 = awkward1.Array([[1.1, 2.2, 3.3], [], [4.4, 5.5]]).layout
    content1 = awkward1.Array(["one", "two", "three", "four", "five"]).layout
    tags = awkward1.layout.Index8(numpy.array([1, 1, 0, 0, 1, 0, 1, 1], dtype=numpy.int8))
    index = awkward1.layout.Index32(numpy.array([0, 1, 0, 1, 2, 2, 4, 3], dtype=numpy.int32))
    array = awkward1.layout.UnionArray8_32(tags, index, [content0, content1])

    array.setidentities()
    assert numpy.asarray(array.identities).tolist() == [[0], [1], [2], [3], [4], [5], [6], [7]]
    assert numpy.asarray(array.content(0).identities).tolist() == [[2], [3], [5]]
    assert numpy.asarray(array.content(1).identities).tolist() == [[0], [1], [4], [7], [6]]

def test_fromiter():
    builder = awkward1.layout.FillableArray()

    builder.integer(0)
    builder.integer(1)
    builder.integer(2)
    builder.beginlist()
    builder.endlist()
    builder.beginlist()
    builder.real(1.1)
    builder.endlist()
    builder.beginlist()
    builder.real(1.1)
    builder.real(2.2)
    builder.endlist()
    builder.beginlist()
    builder.real(1.1)
    builder.real(2.2)
    builder.real(3.3)
    builder.endlist()

    assert awkward1.tolist(builder.snapshot()) == [0, 1, 2, [], [1.1], [1.1, 2.2], [1.1, 2.2, 3.3]]

    assert awkward1.tolist(awkward1.fromiter([0, 1, 2, [], [1.1], [1.1, 2.2], [1.1, 2.2, 3.3]])) == [0, 1, 2, [], [1.1], [1.1, 2.2], [1.1, 2.2, 3.3]]
    assert awkward1.tolist(awkward1.fromiter([0, 1, 2, [], "zero", [1.1], "one", [1.1, 2.2], "two", [1.1, 2.2, 3.3], "three"])) == [0, 1, 2, [], "zero", [1.1], "one", [1.1, 2.2], "two", [1.1, 2.2, 3.3], "three"]
    assert awkward1.tolist(awkward1.fromjson('[0, 1, 2, [], "zero", [1.1], "one", [1.1, 2.2], "two", [1.1, 2.2, 3.3], "three"]')) == [0, 1, 2, [], "zero", [1.1], "one", [1.1, 2.2], "two", [1.1, 2.2, 3.3], "three"]
    assert awkward1.tojson(awkward1.fromjson('[0,1,2,[],"zero",[1.1],"one",[1.1,2.2],"two",[1.1,2.2,3.3],"three"]')) == '[0,1,2,[],"zero",[1.1],"one",[1.1,2.2],"two",[1.1,2.2,3.3],"three"]'



# BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE

import sys
import itertools

import pytest
import numpy

import awkward1

py27 = (sys.version_info[0] < 3)

def test_slice():
    assert repr(awkward1.layout.Slice(3)) == "[3]"
    assert repr(awkward1.layout.Slice(slice(None))) == "[:]"
    assert repr(awkward1.layout.Slice(slice(10))) == "[:10]"
    assert repr(awkward1.layout.Slice(slice(1, 2))) == "[1:2]"
    assert repr(awkward1.layout.Slice(slice(1, None))) == "[1:]"
    assert repr(awkward1.layout.Slice(slice(None, None, 2))) == "[::2]"
    assert repr(awkward1.layout.Slice(slice(1, 2, 3))) == "[1:2:3]"
    if not py27:
        assert repr(awkward1.layout.Slice(Ellipsis)) == "[...]"
    assert repr(awkward1.layout.Slice(numpy.newaxis)) == "[newaxis]"
    assert repr(awkward1.layout.Slice(None)) == "[newaxis]"
    assert repr(awkward1.layout.Slice([1, 2, 3])) == "[array([1, 2, 3])]"
    assert repr(awkward1.layout.Slice(numpy.array([[1, 2], [3, 4], [5, 6]]))) == "[array([[1, 2], [3, 4], [5, 6]])]"
    assert repr(awkward1.layout.Slice(numpy.array([1, 2, 3, 4, 5, 6])[::-2])) == "[array([6, 4, 2])]"
    a = numpy.arange(3*5).reshape(3, 5)[1::, ::-2]
    assert repr(awkward1.layout.Slice(a)) == "[array(" + str(a.tolist()) + ")]"
    a = numpy.arange(3*5).reshape(3, 5)[::-1, ::2]
    assert repr(awkward1.layout.Slice(a)) == "[array(" + str(a.tolist()) + ")]"
    assert repr(awkward1.layout.Slice([True, True, False, False, True])) == "[array([0, 1, 4])]"
    assert repr(awkward1.layout.Slice([[True, True], [False, False], [True, False]])) == "[array([0, 0, 2]), array([0, 1, 0])]"
    assert repr(awkward1.layout.Slice(())) == "[]"
    assert repr(awkward1.layout.Slice((3,))) == "[3]"
    assert repr(awkward1.layout.Slice((3, slice(1, 2, 3)))) == "[3, 1:2:3]"
    assert repr(awkward1.layout.Slice((slice(None), [1, 2, 3]))) == "[:, array([1, 2, 3])]"
    assert repr(awkward1.layout.Slice(([1, 2, 3], slice(None)))) == "[array([1, 2, 3]), :]"
    assert repr(awkward1.layout.Slice((slice(None), [True, True, False, False, True]))) == "[:, array([0, 1, 4])]"
    assert repr(awkward1.layout.Slice((slice(None), [[True, True], [False, False], [True, False]]))) == "[:, array([0, 0, 2]), array([0, 1, 0])]"
    assert repr(awkward1.layout.Slice(([[True, True], [False, False], [True, False]], slice(None)))) == "[array([0, 0, 2]), array([0, 1, 0]), :]"

    with pytest.raises(ValueError):
        awkward1.layout.Slice(numpy.array([1.1, 2.2, 3.3]))
    assert repr(awkward1.layout.Slice(numpy.array(["one", "two", "three"]))) == '[["one", "two", "three"]]'
    with pytest.raises(ValueError):
        awkward1.layout.Slice(numpy.array([1, 2, 3, None, 4, 5]))

    assert repr(awkward1.layout.Slice((123, [[1, 1], [2, 2], [3, 3]]))) == "[array([[123, 123], [123, 123], [123, 123]]), array([[1, 1], [2, 2], [3, 3]])]"
    assert repr(awkward1.layout.Slice(([[1, 1], [2, 2], [3, 3]], 123))) == "[array([[1, 1], [2, 2], [3, 3]]), array([[123, 123], [123, 123], [123, 123]])]"
    assert repr(awkward1.layout.Slice(([[100, 200, 300, 400]], [[1], [2], [3]]))) == "[array([[100, 200, 300, 400], [100, 200, 300, 400], [100, 200, 300, 400]]), array([[1, 1, 1, 1], [2, 2, 2, 2], [3, 3, 3, 3]])]"
    assert repr(awkward1.layout.Slice(([[1], [2], [3]], [[100, 200, 300, 400]]))) == "[array([[1, 1, 1, 1], [2, 2, 2, 2], [3, 3, 3, 3]]), array([[100, 200, 300, 400], [100, 200, 300, 400], [100, 200, 300, 400]])]"

    with pytest.raises(ValueError):
        awkward1.layout.Slice((3, slice(None), [[1], [2], [3]]))
    with pytest.raises(ValueError):
        awkward1.layout.Slice(([[1, 2, 3, 4]], slice(None), [[1], [2], [3]]))
    with pytest.raises(ValueError):
        awkward1.layout.Slice((slice(None), 3, slice(None), [[1], [2], [3]], slice(None)))
    with pytest.raises(ValueError):
        awkward1.layout.Slice((slice(None), [[1, 2, 3, 4]], slice(None), [[1], [2], [3]], slice(None)))
    assert repr(awkward1.layout.Slice((slice(None), 3, [[1], [2], [3]], slice(None)))) == "[:, array([[3], [3], [3]]), array([[1], [2], [3]]), :]"
    assert repr(awkward1.layout.Slice((slice(None), [[1, 2, 3, 4]], [[1], [2], [3]], slice(None)))) == "[:, array([[1, 2, 3, 4], [1, 2, 3, 4], [1, 2, 3, 4]]), array([[1, 1, 1, 1], [2, 2, 2, 2], [3, 3, 3, 3]]), :]"

def test_numpyarray_getitem_bystrides():
    a = numpy.arange(10)
    b = awkward1.layout.NumpyArray(a)
    assert b[3] == a[3]
    assert b[-3] == a[-3]
    assert awkward1.tolist(b[()]) == awkward1.tolist(a[()])
    assert awkward1.tolist(b[slice(None)]) == awkward1.tolist(a[slice(None)])
    assert awkward1.tolist(b[slice(3, 7)]) == awkward1.tolist(a[slice(3, 7)])
    assert awkward1.tolist(b[slice(3, 100)]) == awkward1.tolist(a[slice(3, 100)])
    assert awkward1.tolist(b[slice(-100, 7)]) == awkward1.tolist(a[slice(-100, 7)])
    assert awkward1.tolist(b[slice(3, -3)]) == awkward1.tolist(a[slice(3, -3)])
    assert awkward1.tolist(b[slice(1, 7, 2)]) == awkward1.tolist(a[slice(1, 7, 2)])
    assert awkward1.tolist(b[slice(-8, 7, 2)]) == awkward1.tolist(a[slice(-8, 7, 2)])
    assert awkward1.tolist(b[slice(None, 7, 2)]) == awkward1.tolist(a[slice(None, 7, 2)])
    assert awkward1.tolist(b[slice(None, -3, 2)]) == awkward1.tolist(a[slice(None, -3, 2)])
    assert awkward1.tolist(b[slice(8, None, -1)]) == awkward1.tolist(a[slice(8, None, -1)])
    assert awkward1.tolist(b[slice(8, None, -2)]) == awkward1.tolist(a[slice(8, None, -2)])
    assert awkward1.tolist(b[slice(-2, None, -2)]) == awkward1.tolist(a[slice(-2, None, -2)])

    a = numpy.arange(7*5).reshape(7, 5)
    b = awkward1.layout.NumpyArray(a)

    assert awkward1.tolist(b[()]) == awkward1.tolist(a[()])
    assert awkward1.tolist(b[3]) == awkward1.tolist(a[3])
    assert awkward1.tolist(b[(3, 2)]) == awkward1.tolist(a[3, 2])
    assert awkward1.tolist(b[slice(1, 4)]) == awkward1.tolist(a[slice(1, 4)])
    assert awkward1.tolist(b[(3, slice(1, 4))]) == awkward1.tolist(a[3, slice(1, 4)])
    assert awkward1.tolist(b[(slice(1, 4), 3)]) == awkward1.tolist(a[slice(1, 4), 3])
    assert awkward1.tolist(b[(slice(1, 4), slice(2, None))]) == awkward1.tolist(a[slice(1, 4), slice(2, None)])
    assert awkward1.tolist(b[(slice(None, None, -1), slice(2, None))]) == awkward1.tolist(a[slice(None, None, -1), slice(2, None)])

    assert awkward1.tolist(b[:, numpy.newaxis, :]) == awkward1.tolist(a[:, numpy.newaxis, :])
    assert awkward1.tolist(b[numpy.newaxis, :, numpy.newaxis, :, numpy.newaxis]) == awkward1.tolist(a[numpy.newaxis, :, numpy.newaxis, :, numpy.newaxis])

    if not py27:
        assert awkward1.tolist(b[Ellipsis, 3]) == awkward1.tolist(a[Ellipsis, 3])
        assert awkward1.tolist(b[Ellipsis, 3, 2]) == awkward1.tolist(a[Ellipsis, 3, 2])
        assert awkward1.tolist(b[3, Ellipsis]) == awkward1.tolist(a[3, Ellipsis])
        assert awkward1.tolist(b[3, 2, Ellipsis]) == awkward1.tolist(a[3, 2, Ellipsis])

def test_numpyarray_contiguous():
    a = numpy.arange(10)[8::-2]
    b = awkward1.layout.NumpyArray(a)

    assert awkward1.tolist(b) == awkward1.tolist(a)
    assert awkward1.tolist(b.contiguous()) == awkward1.tolist(a)
    b.become_contiguous()
    assert awkward1.tolist(b) == awkward1.tolist(a)

    a = numpy.arange(7*5).reshape(7, 5)[::-1, ::2]
    b = awkward1.layout.NumpyArray(a)

    assert awkward1.tolist(b) == awkward1.tolist(a)
    assert awkward1.tolist(b.contiguous())
    b.become_contiguous()
    assert awkward1.tolist(b) == awkward1.tolist(a)

def test_numpyarray_getitem_next():
    a = numpy.arange(10)
    b = awkward1.layout.NumpyArray(a)
    c = numpy.array([7, 3, 3, 5])
    assert awkward1.tolist(b[c]) == awkward1.tolist(a[c])
    c = numpy.array([False, False, False, True, True, True, False, True, False, True])
    assert awkward1.tolist(b[c]) == awkward1.tolist(a[c])
    c = numpy.array([], dtype=int)
    assert awkward1.tolist(b[c]) == awkward1.tolist(a[c])

    a = numpy.arange(10*3).reshape(10, 3)
    b = awkward1.layout.NumpyArray(a)
    c = numpy.array([7, 3, 3, 5])
    assert awkward1.tolist(b[c]) == awkward1.tolist(a[c])
    c = numpy.array([False, False, False, True, True, True, False, True, False, True])
    assert awkward1.tolist(b[c]) == awkward1.tolist(a[c])
    c = numpy.array([], dtype=int)
    assert awkward1.tolist(b[c]) == awkward1.tolist(a[c])

    a = numpy.arange(7*5).reshape(7, 5)
    b = awkward1.layout.NumpyArray(a)
    c1 = numpy.array([4, 1, 1, 3])
    c2 = numpy.array([2, 2, 0, 1])
    assert awkward1.tolist(b[c1, c2]) == awkward1.tolist(a[c1, c2])
    c1 = numpy.array([[4, 1], [1, 3], [0, 4]])
    c2 = numpy.array([[2, 2], [0, 1], [1, 3]])
    assert awkward1.tolist(b[c1, c2]) == awkward1.tolist(a[c1, c2])
    c = numpy.array([False, False, True, True, False, True, True])
    assert awkward1.tolist(b[c]) == awkward1.tolist(a[c])
    c = (a % 2 == 0)   # two dimensional
    assert awkward1.tolist(b[c]) == awkward1.tolist(a[c])
    c = numpy.array([], dtype=int)
    assert awkward1.tolist(b[c]) == awkward1.tolist(a[c])
    c1 = numpy.array([], dtype=int)
    c2 = numpy.array([], dtype=int)
    assert awkward1.tolist(b[c1, c2]) == awkward1.tolist(a[c1, c2])

    a = numpy.arange(7*5).reshape(7, 5)
    b = awkward1.layout.NumpyArray(a)
    c = numpy.array([2, 0, 0, 1])
    assert awkward1.tolist(b[1:4, c]) == awkward1.tolist(a[1:4, c])
    assert awkward1.tolist(b[c, 1:4]) == awkward1.tolist(a[c, 1:4])
    assert awkward1.tolist(b[1:4, numpy.newaxis, c]) == awkward1.tolist(a[1:4, numpy.newaxis, c])
    assert awkward1.tolist(b[c, numpy.newaxis, 1:4]) == awkward1.tolist(a[c, numpy.newaxis, 1:4])
    assert awkward1.tolist(b[1:4, numpy.newaxis, numpy.newaxis, c]) == awkward1.tolist(a[1:4, numpy.newaxis, numpy.newaxis, c])
    assert awkward1.tolist(b[c, numpy.newaxis, numpy.newaxis, 1:4]) == awkward1.tolist(a[c, numpy.newaxis, numpy.newaxis, 1:4])
    if not py27:
        assert awkward1.tolist(b[Ellipsis, c]) == awkward1.tolist(a[Ellipsis, c])
        assert awkward1.tolist(b[c, Ellipsis]) == awkward1.tolist(a[c, Ellipsis])



# BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE

import sys
import itertools
import collections

import pytest
import numpy

import awkward1
import awkward1.behavior.string

py27 = (sys.version_info[0] < 3)

def test_fromnumpy():
    a = numpy.arange(2*3*5).reshape((2, 3, 5))
    b = awkward1.fromnumpy(a)
    assert awkward1.tolist(a) == awkward1.tolist(b)

def test_highlevel():
    a = awkward1.Array([[1.1, 2.2, 3.3], [], [4.4, 5.5], [6.6], [7.7, 8.8, 9.9]])
    assert repr(a) == "<Array [[1.1, 2.2, 3.3], ... [7.7, 8.8, 9.9]] type='5 * var * float64'>"
    assert str(a) == "[[1.1, 2.2, 3.3], [], [4.4, 5.5], [6.6], [7.7, 8.8, 9.9]]"

    b = awkward1.Array(numpy.arange(100, dtype=numpy.int32))
    assert repr(b) == "<Array [0, 1, 2, 3, 4, ... 95, 96, 97, 98, 99] type='100 * int32'>"
    assert str(b) == "[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, ... 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99]"

    c = awkward1.Array('[{"one": 3.14, "two": [1.1, 2.2]}, {"one": 99.9, "two": [-3.1415926]}]')
    assert repr(c) == "<Array [{one: 3.14, two: [1.1, ... -3.14]}] type='2 * {\"one\": float64, \"two\": va...'>"
    assert str(c) == "[{one: 3.14, two: [1.1, 2.2]}, {one: 99.9, two: [-3.14]}]"

class Dummy(awkward1.highlevel.Array):
    pass

def test_string1():
    a = awkward1.Array(numpy.array([ord(x) for x in "hey there"], dtype=numpy.uint8))
    a.__class__ = awkward1.behavior.string.CharBehavior
    assert str(a) == str(b"hey there")
    assert repr(a) == repr(b"hey there")

def test_string2():
    content = awkward1.layout.NumpyArray(numpy.array([ord(x) for x in "heythere"], dtype=numpy.uint8))
    listoffsetarray = awkward1.layout.ListOffsetArray64(awkward1.layout.Index64(numpy.array([0, 3, 3, 8])), content)
    a = awkward1.Array(listoffsetarray)

    assert isinstance(a, awkward1.Array)
    assert not isinstance(a, awkward1.behavior.string.StringBehavior)
    assert awkward1.tolist(a) == [[104, 101, 121], [], [116, 104, 101, 114, 101]]

    assert repr(a.type) == "3 * var * uint8"
    assert repr(a[0].type) == "3 * uint8"
    assert repr(a[1].type) == "0 * uint8"
    assert repr(a[2].type) == "5 * uint8"

    assert repr(a) == "<Array [[104, 101, 121], ... 101, 114, 101]] type='3 * var * uint8'>"
    assert str(a) == "[[104, 101, 121], [], [116, 104, 101, 114, 101]]"
    assert repr(a[0]) == "<Array [104, 101, 121] type='3 * uint8'>"
    assert repr(a[1]) == "<Array [] type='0 * uint8'>"
    assert repr(a[2]) == "<Array [116, 104, 101, 114, 101] type='5 * uint8'>"

    a = awkward1.Array(listoffsetarray.astype(awkward1.string))
    assert isinstance(a, awkward1.Array)
    assert awkward1.tolist(a) == ['hey', '', 'there']

    assert repr(a.type) == "3 * string"
    assert repr(a[0].type) == "3 * utf8"
    assert repr(a[1].type) == "0 * utf8"
    assert repr(a[2].type) == "5 * utf8"

    if py27:
        assert repr(a) == "<Array [u'hey', u'', u'there'] type='3 * string'>"
        assert repr(a[0]) == "u'hey'"
        assert repr(a[1]) == "u''"
        assert repr(a[2]) == "u'there'"
    else:
        assert repr(a) == "<Array ['hey', '', 'there'] type='3 * string'>"
        assert repr(a[0]) == "'hey'"
        assert repr(a[1]) == "''"
        assert repr(a[2]) == "'there'"

def test_accepts():
    dressed1 = awkward1.layout.ListType(awkward1.layout.PrimitiveType("float64"), {"__class__": "Dummy"})
    content = awkward1.layout.NumpyArray(numpy.array([1.1, 2.2, 3.3, 4.4, 5.5], dtype=numpy.float64))
    listoffsetarray = awkward1.layout.ListOffsetArray64(awkward1.layout.Index64(numpy.array([0, 3, 3, 5])), content).astype(dressed1)

    dressed2 = awkward1.layout.PrimitiveType("float64", {"__class__": "Dummy"})
    with pytest.raises(ValueError):
        awkward1.layout.ListOffsetArray64(awkward1.layout.Index64(numpy.array([0, 3, 3, 5])), content).astype(dressed2)

class D(awkward1.highlevel.Array):
    @staticmethod
    def typestr(baretype, parameters):
        return "D[{0}]".format(baretype)

def test_type_propagation():
    array = awkward1.Array([[{"one": 1, "two": [1.0, 1.1]}, {"one": 2, "two": [2.0]}, {"one": 3, "two": [3.0, 3.1, 3.2]}], [], [{"one": 4, "two": []}, {"one": 5, "two": [5.0, 5.1, 5.2, 5.3]}]])
    assert awkward1.tolist(array) == [[{"one": 1, "two": [1.0, 1.1]}, {"one": 2, "two": [2.0]}, {"one": 3, "two": [3.0, 3.1, 3.2]}], [], [{"one": 4, "two": []}, {"one": 5, "two": [5.0, 5.1, 5.2, 5.3]}]]
    assert repr(array.type) in ('3 * var * {"one": int64, "two": var * float64}', '3 * var * {"two": var * float64, "one": int64}')

    dfloat64 = awkward1.layout.PrimitiveType("float64", {"__class__": "D", "__typestr__": "D[float64]"})
    dvarfloat64 = awkward1.layout.ListType(dfloat64, {"__class__": "D", "__typestr__": "D[var * D[float64]]"})
    dint64 = awkward1.layout.PrimitiveType("int64", {"__class__": "D", "__typestr__": "D[int64]"})
    drec = awkward1.layout.RecordType(collections.OrderedDict([("one", dint64), ("two", dvarfloat64)]), {"__class__": "D", "__typestr__": "D[{\"one\": D[int64], \"two\": D[var * D[float64]]}]"})
    dvarrec = awkward1.layout.ListType(drec, {"__class__": "D", "__typestr__": "D[var * D[{\"one\": D[int64], \"two\": D[var * D[float64]]}]]"})

    array = awkward1.Array(array.layout.astype(dvarrec))

    assert array.layout.type == dvarrec
    assert array.layout.content.type == drec
    assert array.layout.content.field("one").type == dint64
    assert array.layout.content.field("two").type == dvarfloat64
    assert array.layout.content.field("two").content.type == dfloat64

    assert array.layout[-1].type == drec
    assert array.layout[-1]["one"].type == dint64
    assert array.layout[-1]["two"].type == dvarfloat64
    assert array.layout[-1]["two"][1].type == dfloat64
    assert array.layout[-1, "one"].type == dint64
    assert array.layout[-1, "two"].type == dvarfloat64
    assert array.layout[-1, "two", 1].type == dfloat64
    assert array.layout["one", -1].type == dint64
    assert array.layout["two", -1].type == dvarfloat64
    assert array.layout["two", -1, 1].type == dfloat64

    assert array.layout[1:].type == dvarrec
    assert array.layout[1:, "one"].type == awkward1.layout.ListType(dint64)
    assert array.layout["one", 1:].type == awkward1.layout.ListType(dint64)

    assert array.layout[[2, 1]].type == dvarrec
    assert array.layout[[2, 1], "one"].type == awkward1.layout.ListType(dint64)

    array2 = awkward1.layout.NumpyArray(numpy.arange(2*3*5, dtype=numpy.int64).reshape(2, 3, 5)).astype(awkward1.layout.RegularType(awkward1.layout.RegularType(dint64, 5), 3))

    assert repr(array2.type) == "3 * 5 * D[int64]"
    assert repr(array2[0].type) == "5 * D[int64]"
    assert repr(array2[0, 0].type) == "D[int64]"
    assert array2[-1, -1, -1] == 29



# BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE

import sys
import numbers
try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable

import numpy

import awkward1._util
import awkward1.layout
import awkward1._numpy
import awkward1.operations.convert

def isna(array):
    import awkward1.highlevel

    def apply(layout):
        if isinstance(layout, awkward1.layout.EmptyArray):
            return apply(awkward1.layout.NumpyArray(numpy.array([])))

        elif isinstance(layout, (awkward1.layout.IndexedArray32, awkward1.layout.IndexedArrayU32, awkward1.layout.IndexedArray64)):
            return apply(layout.project())

        elif isinstance(layout, (awkward1.layout.UnionArray8_32, awkward1.layout.UnionArray8_U32, awkward1.layout.UnionArray8_64)):
            contents = [apply(layout.project(i)) for i in range(len(layout))]
            out = numpy.empty(len(layout), dtype=numpy.bool_)
            tags = numpy.asarray(layout.tags)
            for tag, content in enumerate(contents):
                out[tags == tag] = content
            return out

        elif isinstance(layout, (awkward1.layout.IndexedOptionArray32, awkward1.layout.IndexedOptionArray64)):
            index = numpy.asarray(layout.index)
            return (index < 0)

        else:
            return numpy.zeros(len(layout), dtype=numpy.bool_)

    return awkward1.highlevel.Array(apply(awkward1.operations.convert.tolayout(array, allowrecord=False)))

def notna(array):
    return ~isna(array)

@awkward1._numpy.implements(numpy.size)
def size(array, axis=None):
    if axis is not None and axis < 0:
        raise NotImplementedError("ak.size with axis < 0")

    def recurse(layout, axis, sizes):
        if isinstance(layout, awkward1._util.unknowntypes):
            pass
        elif isinstance(layout, awkward1._util.indexedtypes):
            recurse(layout.content, axis, sizes)
        elif isinstance(layout, awkward1._util.uniontypes):
            compare = None
            for x in layout.contents:
                inner = []
                recurse(x, axis, inner)
                if compare is None:
                    compare = inner
                elif compare != inner:
                    raise ValueError("ak.size is ambiguous due to union of different sizes")
            sizes.extend(compare)
        elif isinstance(layout, awkward1._util.optiontypes):
            return recurse(layout.content, axis, sizes)
        elif isinstance(layout, awkward1._util.listtypes):
            if isinstance(layout, awkward1.layout.RegularArray):
                sizes.append(layout.size)
            else:
                sizes.append(None)
            if axis is None:
                recurse(layout.content, axis, sizes)
            elif axis > 0:
                recurse(layout.content, axis - 1, sizes)
        elif isinstance(layout, awkward1._util.recordtypes):
            compare = None
            for x in layout.contents:
                inner = []
                recurse(x, axis, inner)
                if compare is None:
                    compare = inner
                elif compare != inner:
                    raise ValueError("ak.size is ambiguous due to record of different sizes")
            sizes.extend(compare)
        elif isinstance(layout, awkward1.layout.NumpyArray):
            if axis is None:
                sizes.extend(numpy.asarray(layout).shape[1:])
            else:
                sizes.extend(numpy.asarray(layout).shape[1:axis + 2])
        else:
            raise AssertionError("unrecognized Content type")

    layout = awkward1.operations.convert.tolayout(array, allowrecord=False)
    layout = awkward1.layout.RegularArray(layout, len(layout))

    sizes = []
    recurse(layout, axis, sizes)

    if axis is None:
        out = 1
        for size in sizes:
            if size is None:
                raise ValueError("ak.size is ambiguous due to variable-length arrays (try ak.flatten to remove structure or ak.tonumpy to force regularity, if possible)")
            else:
                out *= size
        return out
    else:
        if sizes[-1] is None:
            raise ValueError("ak.size is ambiguous due to variable-length arrays at axis {0} (try ak.flatten to remove structure or ak.tonumpy to force regularity, if possible)".format(axis))
        else:
            return sizes[-1]

@awkward1._numpy.implements(numpy.atleast_1d)
def atleast_1d(*arrays):
    return numpy.atleast_1d(*[awkward1.operations.convert.tonumpy(x) for x in arrays])

@awkward1._numpy.implements(numpy.concatenate)
def concatenate(arrays, axis=0):
    import awkward1.highlevel

    if axis != 0:
        raise NotImplementedError("axis={0}".format(axis))

    contents = [awkward1.operations.convert.tolayout(x, allowrecord=False) for x in arrays]

    tags = numpy.empty(sum(len(x) for x in contents), dtype=numpy.int8)
    index = numpy.empty(len(tags), dtype=numpy.int64)
    start = 0
    for tag, x in enumerate(contents):
        tags[start : start + len(x)] = tag
        index[start : start + len(x)] = numpy.arange(len(x), dtype=numpy.int64)
        start += len(x)

    tags = awkward1.layout.Index8(tags)
    index = awkward1.layout.Index64(index)
    out = awkward1.layout.UnionArray8_64(tags, index, contents)
    return awkward1.highlevel.Array(out)

@awkward1._numpy.implements(numpy.where)
def where(condition, *args):
    import awkward1.highlevel

    condition = awkward1.operations.convert.tonumpy(condition)

    if len(args) == 0:
        out = numpy.nonzero(condition)
        return tuple(awkward1.highlevel.Array(x) for x in out)

    elif len(args) == 1:
        raise ValueError("either both or neither of x and y should be given")

    elif len(args) == 2:
        if len(condition.shape) != 1:
            raise NotImplementedError("FIXME: ak.where(condition, x, y) where condition is not 1-d")

        x = awkward1.operations.convert.tolayout(args[0], allowrecord=False)
        y = awkward1.operations.convert.tolayout(args[1], allowrecord=False)

        tags = (condition == 0)
        assert tags.itemsize == 1
        index = numpy.empty(len(tags), dtype=numpy.int64)
        index = numpy.arange(len(condition), dtype=numpy.int64)

        tags = awkward1.layout.Index8(tags.view(numpy.int8))
        index = awkward1.layout.Index64(index)
        # FIXME: call "simplify" when it exists
        return awkward1.highlevel.Array(awkward1.layout.UnionArray8_64(tags, index, [x, y]))

    else:
        raise TypeError("where() takes from 1 to 3 positional arguments but {0} were given".format(len(args) + 1))

__all__ = [x for x in list(globals()) if not x.startswith("_") and x not in ("numbers", "Iterable", "numpy", "awkward1")]



# BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE

import sys
import os
import json

import pytest
import numpy

import awkward1

def test_unknown():
    a = awkward1.fromjson("[[], [], []]").layout
    assert awkward1.tolist(a) == [[], [], []]
    assert str(awkward1.typeof(a)) == "var * unknown"
    assert awkward1.typeof(a) == awkward1.layout.ListType(awkward1.layout.UnknownType())
    assert not awkward1.typeof(a) == awkward1.layout.PrimitiveType("float64")

    a = awkward1.fromjson("[[], [[], []], [[], [], []]]").layout
    assert awkward1.tolist(a) == [[], [[], []], [[], [], []]]
    assert str(awkward1.typeof(a)) == "var * var * unknown"
    assert awkward1.typeof(a) == awkward1.layout.ListType(awkward1.layout.ListType(awkward1.layout.UnknownType()))

    a = awkward1.layout.FillableArray()
    a.beginlist()
    a.endlist()
    a.beginlist()
    a.endlist()
    a.beginlist()
    a.endlist()
    assert awkward1.tolist(a) == [[], [], []]
    assert str(awkward1.typeof(a)) == "var * unknown"
    assert awkward1.typeof(a) == awkward1.layout.ListType(awkward1.layout.UnknownType())
    assert not awkward1.typeof(a) == awkward1.layout.PrimitiveType("float64")

    a = a.snapshot()
    assert awkward1.tolist(a) == [[], [], []]
    assert str(awkward1.typeof(a)) == "var * unknown"
    assert awkward1.typeof(a) == awkward1.layout.ListType(awkward1.layout.UnknownType())
    assert not awkward1.typeof(a) == awkward1.layout.PrimitiveType("float64")

def test_getitem():
    a = awkward1.fromjson("[[], [[], []], [[], [], []]]")
    assert awkward1.tolist(a[2]) == [[], [], []]

    assert awkward1.tolist(a[2, 1]) == []
    with pytest.raises(ValueError) as excinfo:
        a[2, 1, 0]
    assert str(excinfo.value) == "in ListArray64 attempting to get 0, index out of range"
    assert awkward1.tolist(a[2, 1][()]) == []
    with pytest.raises(ValueError) as excinfo:
        a[2, 1][0]
    assert str(excinfo.value) == "in EmptyArray attempting to get 0, index out of range"
    assert awkward1.tolist(a[2, 1][100:200]) == []
    assert awkward1.tolist(a[2, 1, 100:200]) == []
    assert awkward1.tolist(a[2, 1][numpy.array([], dtype=int)]) == []
    assert awkward1.tolist(a[2, 1, numpy.array([], dtype=int)]) == []
    with pytest.raises(ValueError) as excinfo:
        a[2, 1, numpy.array([0], dtype=int)]
    assert str(excinfo.value) == "in ListArray64 attempting to get 0, index out of range"
    with pytest.raises(ValueError) as excinfo:
        a[2, 1][100:200, 0]
    assert str(excinfo.value) == "in EmptyArray, too many dimensions in slice"
    with pytest.raises(ValueError) as excinfo:
        a[2, 1][100:200, 200:300]
    assert str(excinfo.value) == "in EmptyArray, too many dimensions in slice"
    with pytest.raises(ValueError) as excinfo:
        a[2, 1][100:200, numpy.array([], dtype=int)]
    assert str(excinfo.value) == "in EmptyArray, too many dimensions in slice"

    assert awkward1.tolist(a[1:, 1:]) == [[[]], [[], []]]
    with pytest.raises(ValueError) as excinfo:
        a[1:, 1:, 0]
    assert str(excinfo.value) == "in ListArray64 attempting to get 0, index out of range"

numba = pytest.importorskip("numba")
def test_numba():
    a = awkward1.fromjson("[[], [[], []], [[], [], []]]").layout

    @numba.njit
    def f1(q):
        return q[2, 1]
    assert awkward1.tolist(f1(a)) == []

    @numba.njit
    def f2(q):
        return q[2, 1][()]
    assert awkward1.tolist(f2(a)) == []

    @numba.njit
    def f3(q):
        return q[2, 1][100:200]
    assert awkward1.tolist(f3(a)) == []

    @numba.njit
    def f4(q):
        return q[2, 1, 0]
    with pytest.raises(numba.errors.TypingError):
        f4(a)

    @numba.njit
    def f5(q):
        return q[2, 1, 100:200]
    assert awkward1.tolist(f5(a)) == []

    @numba.njit
    def f6a(q):
        return q[2, 1, 100:200, 0]
    with pytest.raises(numba.errors.TypingError):
        f6a(a)

    @numba.njit
    def f6b(q):
        return q[2, 1, 100:200][0]
    with pytest.raises(numba.errors.TypingError):
        f6b(a)

    @numba.njit
    def f7a(q):
        return q[2, 1, 100:200, 200:300]
    with pytest.raises(numba.errors.TypingError):
        f7a(a)

    @numba.njit
    def f7b(q):
        return q[2, 1, 100:200][200:300]
    assert awkward1.tolist(f7b(a)) == []

    @numba.njit
    def f7c(q):
        return q[2, 1, 100:200][()]
    assert awkward1.tolist(f7c(a)) == []

    @numba.njit
    def f8a(q):
        return q[2, 1, 100:200, numpy.array([], dtype=numpy.int64)]
    with pytest.raises(numba.errors.TypingError):
        f8a(a)

    @numba.njit
    def f8b(q, z):
        return q[2, 1, z]
    assert awkward1.tolist(f8b(a, numpy.array([], dtype=int))) == []

    @numba.njit
    def f8c(q, z):
        return q[2, 1, z, z]
    with pytest.raises(numba.errors.TypingError):
        f8c(a, numpy.array([], dtype=int))

    @numba.njit
    def f8d(q, z):
        return q[2, 1, z][()]
    assert awkward1.tolist(f8d(a, numpy.array([], dtype=int))) == []



# BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE

import sys

import pytest
import numpy
numba = pytest.importorskip("numba")

import awkward1
awkward1_numba_util = pytest.importorskip("awkward1._numba.util")

py27 = (sys.version_info[0] < 3)

def test_slice_utils():
    a = numpy.array([1, 2, 3])
    b = numpy.array([[4], [5], [6], [7]])
    c = 999

    assert [x.tolist() for x in numpy.broadcast_arrays(a, b)] == [x.tolist() for x in awkward1_numba_util.broadcast_arrays((a, b))]
    assert [x.tolist() for x in numpy.broadcast_arrays(b, c)] == [x.tolist() for x in awkward1_numba_util.broadcast_arrays((b, c))]
    assert [x.tolist() for x in numpy.broadcast_arrays(c, a)] == [x.tolist() for x in awkward1_numba_util.broadcast_arrays((c, a))]
    assert [x.tolist() for x in numpy.broadcast_arrays(a, b, c)] == [x.tolist() for x in awkward1_numba_util.broadcast_arrays((a, b, c))]

    if not py27:
        assert awkward1_numba_util.broadcast_arrays(("hello", a))[0] == "hello"
        assert awkward1_numba_util.broadcast_arrays((a, "hello"))[1] == "hello"
        assert awkward1_numba_util.broadcast_arrays(("hello", c)) == ("hello", c)
        assert awkward1_numba_util.broadcast_arrays((c, "hello")) == (c, "hello")
        assert [x.tolist() for x in numpy.broadcast_arrays(a, b)] == [x.tolist() for x in awkward1_numba_util.broadcast_arrays(("hello", a, b))[1:]]

    assert [x.tolist() for x in awkward1_numba_util.regularize_slice((numpy.array([0, 1, 2, 3]), numpy.array([[True, False], [False, True], [True, True], [False, False]])))]

starts  = numpy.array([0, 3, 3, 5, 6])
stops   = numpy.array([3, 3, 5, 6, 10])
content = numpy.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
array   = awkward1.layout.ListArray64(awkward1.layout.Index64(starts),
                                      awkward1.layout.Index64(stops),
                                      awkward1.layout.NumpyArray(content))

def test_boxing():
    assert awkward1.tolist(array) == [[0.0, 1.1, 2.2], [], [3.3, 4.4], [5.5], [6.6, 7.7, 8.8, 9.9]]
    @numba.njit
    def f1(q):
        return q
    assert awkward1.tolist(f1(array)) == [[0.0, 1.1, 2.2], [], [3.3, 4.4], [5.5], [6.6, 7.7, 8.8, 9.9]]

def test_simple():
    @numba.njit
    def f1(q):
        return q[2]

    assert awkward1.tolist(f1(array)) == [3.3, 4.4]

    @numba.njit
    def f2(q):
        return q[2:4]

    assert awkward1.tolist(f2(array)) == [[3.3, 4.4], [5.5]]

    @numba.njit
    def f3(q):
        return q[()]

    assert awkward1.tolist(f3(array)) == [[0.0, 1.1, 2.2], [], [3.3, 4.4], [5.5], [6.6, 7.7, 8.8, 9.9]]

def test_array():
    @numba.njit
    def f1(q):
        return q[[2, 0, 0, 1],]

    assert awkward1.tolist(f1(array)) == [[3.3, 4.4], [0.0, 1.1, 2.2], [0.0, 1.1, 2.2], []]



# BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE

import sys

import pytest
import numpy

import awkward1

class Point(awkward1.Record):
  def __repr__(self):
    return "<{} {}>".format(self["x"], self["y"])

def test():
    array = awkward1.Array([[{"x": 1, "y": [1.1]}, {"x": 2, "y": [2.0, 0.2]}], [], [{"x": 3, "y": [3.0, 0.3, 3.3]}]])
    array.layout.content.setparameter("__class__", "Point")
    array.layout.content.setparameter("__typestr__", "P")
    awkward1.classes["Point"] = Point
    assert repr(array[0, 0]) == "<1 [1.1]>"
    assert repr(array[0]) == "<Array [<1 [1.1]>, <2 [2, 0.2]>] type='2 * P'>"
    assert repr(array) == "<Array [[<1 [1.1]>, ... <3 [3, 0.3, 3.3]>]] type='3 * var * P'>"
